(*********************************************************************
**********************************************************************
****                                                              ****
****                        COMM.INC                              ****
****                                                              ****
****         Interrupt serviced serial communications.            ****
****                                                              ****
**********************************************************************
**********************************************************************)


Unit Comm;


INTERFACE


Uses Dos, Crt, Global, Sounds, Math, Strings, Video, Time, Key, Disk;


Const
  MaxComPort = 2;
  ComPortAddr : Array [ 1..MaxComPort ] of Word = ( $3F8, $2F8 );
  { use these constants for accessing ports after the base address }
  IntrEnableOfs = 1;
  BaudDivOfs    = 1;
  IntrIDOfs     = 2;
  LineCtrlOfs   = 3;
  ModemCtrlOfs  = 4;
  LineStatOfs   = 5;
  ModemStatOfs  = 6;
  { masks for reading status registers }
  RXReadyMask   = $01;
  THRReadyMask  = $20;  { Transmit Holding Register }
  TSRReadyMask  = $40;  { Transmit Shift Register }
  ComErrMask    = $0E;
  ComIdleMask   = $6F;
  ComRXMask     = $0F;
  { low level IO error codes returned by this module }
  ComPortError       = 126;
  RxCharTimeoutErr   = 127;
  TxCharTimeoutErr   = 128;
  RxUARTParityErr    = 129;
  RxUARTOverrunErr   = 130;
  RxUARTFramingErr   = 131;
  { misc }
  SizeOfSniffBuffer = 255;
  ComDefaultTimeout = 4000;
  SnifferFileSpec = '\SNIFFER.LOG';

Type
  ComRxBufferType = Array[ 0..255 ] Of Byte;
  ComStatStrType = String[ 16 ];
  ComSetupStrType = String[ 10 ];
  ComGenlStrType = String[ 80 ];

Var
  ComInPtr : Array [ 1..MaxComPort ] of Byte;   { rx rcv buffer point }
  ComOutPtr : Array [ 1..MaxComPort ] of Byte;  { rx read buffer point }
  ComRxBuffer : Array [ 1..MaxComPort ] of ^ComRxBufferType;  { rx buffer }
  ComRXErr : Array [ 1..MaxComPort ] of Byte;   { rx UART errors }
  ComISR : Array [ 1..MaxComPort ] of Boolean;  { port is intr. served }
  ComPortSetupSave : Array [ 1..MaxComPort ] of ComSetupStrType;
  ComExitSave : Pointer;    { used to unload com rx ISR }
  ComPortReady : Array[ 1..MaxComPort ] of Boolean;  { port is configured }
  ComResult : Integer;   { IO result for serial communications system }
  SniffBuffer : String[ SizeOfSniffBuffer ];  { sniffer memory buffer }
  ErrorCapture : Boolean;  { force next capture to be error capture }
  PrevPort : Byte;   { current port being sniffed }
  SniffLineLength  : Byte;  { length of current line in sniffer file }
  SniffOverflow : Boolean;  { buffer overflowed since last flushed }
  SniffSize : LongInt;  { current sniffer file size }
  SniffBufferMax : Byte;  { largest buffer accumulated before flushing }
  SniffLog : File of Char;
  SnifferOpen : Boolean;
  Com :
    Record
      SniffMax : LongInt;       { size of sniffer file }
      SnifferRecycle : Boolean;    { sniffer wraps through capture sequence }
      Timeout : Array [ 1..MaxComPort ] of Integer;  { RX/TX timeout }
    { ignore access to selected ports for simulated communications }
      Fake : Array [ 1..MaxComPort ] of Boolean;
      AllFake : Boolean;  { indicates all ports set to fake }
      View : Boolean;            { enables communications monitor window }
     End;

{ Half duplex is rx and tx tied together, full is separate rx and tx. }
{ Note that to support half duplex, the application program must call }
{ ClearHalfDuplex every time data direction is change from tx to rx. }

Function  NextSnifferFileName : ComGenlStrType;
Procedure SniffWrite( Inln : ComGenlStrType );
Procedure SniffWriteLn( InLn : ComGenlStrType );
Procedure SniffFlush;
Procedure CheckPortChange( PortNum:Byte );
Function  ComPortExists( PortNum:Byte ) : Boolean;
Procedure DTROff( PortNum:Byte );
Procedure DTROn( PortNum:Byte );
Procedure RTSOff( PortNum:Byte );
Procedure RTSOn( PortNum:Byte );
Function  ComInit( PortNum, Baud : Integer;  Parity : Char;
                   DataBits, StopBits : Integer;  ISR : Boolean ) : Integer;
Function  ComOutReady( PortNum : Byte ) : Boolean;
Procedure ClearComOut( PortNum : Byte );
Procedure ClearHalfDuplex( PortNum : Byte );
Procedure ComOut( PortNum, B : Byte );
Function  ComInReady( PortNum : Byte ) : Boolean;
Procedure ClearComIn( PortNum : Byte );
Function  ComIn( PortNum : Byte ) : Byte;
Function  LineOK( PortNum, EchoByte : Byte ) : Boolean;
Procedure ComInq( PortNum : Integer;  Var Baud : Integer;
                  Var Parity : Char;  Var DataBits, StopBits : Integer;
                  Var Duplex : Char;  Var ISR : Boolean );
Function  ComSetupStr( ComPort : Byte ) : ComSetupStrType;
Function  ComStatStr( PortNum : Byte ) : ComStatStrType;


IMPLEMENTATION


{----------------------------------------------------------------------}
{ Used by sniffer to find next available capture file name.            }
{----------------------------------------------------------------------}

Function NextSnifferFileName : ComGenlStrType;
  Var  CurrentFile:ComGenlStrType;  X, FindResult, FileCount:Integer;
       WorkFile:File;  Capture:ComGenlStrType;
  Begin
    { finds the first unused file name in \CAPTURE0.LOG .. \CAPTURE9.LOG
      sets SnifferNextFile, then erases next file in circular sequence }
    If ErrorCapture Then Capture := '\CAPERR' Else Capture := '\CAPTURE';
    FileCount := -1;  FindResult := 0;
    While  ( FileCount < 9 )  and  ( FindResult = 0 ) do
      Begin   { until last file, or doesn't exist }
        Inc( FileCount );
        CurrentFile := Capture + Chr( FileCount + 48 ) + '.LOG';
        Assign( WorkFile, CurrentFile );
        {$I-} Reset( WorkFile ); {$I+}  FindResult := IOResult;
        {$I-} Close( WorkFile ); {$I+}  X := IOResult;
      End;
    If FindResult = 0 Then
      Begin
        FileCount := 0;  CurrentFile := Capture + '0.LOG';
      End; { circular }
    NextSnifferFileName := CurrentFile;
    FileCount := ( FileCount + 1 ) mod 10;  { find next in sequence }
    CurrentFile := Capture + Chr( FileCount + 48 ) + '.LOG';
    Assign( WorkFile, CurrentFile );
    {$I-} Erase( WorkFile ); {$I+} X := IOResult;
  End;

{----------------------------------------------------------------------}
{ Write characters to sniffer buffer.                                  }
{----------------------------------------------------------------------}

Procedure SniffWrite( Inln : ComGenlStrType );
  Var  BuffLoop : Byte;
  Begin
    If ( Com.SniffMax = 0 ) and ( not Com.View ) Then Exit;
    If Length( SniffBuffer ) + Length( InLn ) > SizeOfSniffBuffer Then
      Begin
        InLn[ 0 ] := Chr( SizeOfSniffBuffer - Length( SniffBuffer ) );
        SniffOverFlow := True;
      End;
    If InLn <> '' Then
      Begin
        Move( InLn[ 1 ], SniffBuffer[ Length( SniffBuffer ) + 1 ],
              Ord( InLn[ 0 ] ) );
        SniffBuffer[ 0 ] := Chr( Length( SniffBuffer ) + Length( InLn ) );
      End;
    If Length( SniffBuffer ) > SniffBufferMax Then
      SniffBufferMax := Length( SniffBuffer );
  End;

{----------------------------------------------------------------------}
{ Write a line to the sniffer.                                         }
{----------------------------------------------------------------------}

Procedure SniffWriteLn( InLn : ComGenlStrType );
  Begin
    SniffWrite( NL + InLn + NL );
  End;

{----------------------------------------------------------------------}
{ Flush sniffer memory buffer to disk.                                 }
{----------------------------------------------------------------------}

Procedure SniffFlush;
  Var  SC1, SC2 : Char;  SFResultStr : String[10];
       CLoop, SFResult : Integer;  NewName : String[40];
  BEGIN
    If ( Com.SniffMax = 0 ) or Com.AllFake then Exit;
    {$I-}
    SFResult := 0;
    If not SnifferOpen Then
      Begin
        SniffSize := 0;
        Assign( SniffLog, SnifferFileSpec );
        Reset( SniffLog );  SFResult := IOResult;
        If SFResult = FileNotFound Then  { create }
          Begin  ReWrite( SniffLog );  SFResult := IOResult;  End;
        If SFResult = 0 Then
          Begin  { open ok }
            SniffSize := FileSize( SniffLog );
            If SniffSize < Com.SniffMax Then  { file still has space }
              Begin  Seek( SniffLog, SniffSize );  SFResult := IOResult;  End
            Else  { file at limit }
              If Com.SnifferRecycle Then
                Begin  { wrap to capture if applicable }
                  NewName := NextSnifferFileName;
                  Close( SniffLog );  SFResult := IOResult;
                  Rename( SniffLog, NewName );  SFResult := IOResult;
                  Assign( SniffLog, SnifferFileSpec );
                  ReWrite( SniffLog );  SFResult := IOResult;
                  AuditLog( 'SNIFFER:  wrapped to ' + NewName );
                  SniffSize := 0;
                End;  { wrap }
          End;  { open ok }
        SnifferOpen := SFResult = 0;
      End;
    If SFResult = 0 Then
      Begin  { open ok }
        CLoop := 1;
        While ( SFResult = 0 ) and ( CLoop <= Length( SniffBuffer ) ) do
          Begin  { write buffer }
            Write( SniffLog, SniffBuffer[ CLoop ] );  SFResult := IOResult;
            Inc( SniffLineLength );  Inc( SniffSize );
            If ( CLoop > 1 ) and
               ( Copy( SniffBuffer, CLoop, 2 ) = NL ) Then
              SniffLineLength := 0;
            If ( SFResult = 0 ) and
               ( SniffLineLength >= 65 ) and
               ( SniffBuffer[ CLoop ] = ' ' ) Then
              Begin  { wrap line }
                SniffLineLength := 0;  SC1 := #13;  SC2 := #10;
                Write( SniffLog, SC1, SC2 );  SFResult := IOResult;
                SniffSize := SniffSize + 2;
              End;  { wrap line }
            Inc( CLoop );
          End;  { write buffer }
        If SniffOverflow Then
          Begin  { note overflow }
            SniffBuffer := NL + '*** DATA CLIPPED ***' + NL;
            CLoop := 1;
            While ( SFResult = 0 ) and ( CLoop <= Length( SniffBuffer ) ) do
              Begin
                Write( SniffLog, SniffBuffer[ CLoop ] );
                SFResult := IOResult;
                Inc( CLoop );  Inc( SniffSize );
              End;
            SniffOverflow := False;  SniffLineLength := 0;
          End;  { note overflow }
      End;  { open ok }
    If SFResult <> 0 then
      Begin
        Str( SFResult, SFResultStr );
        ErrorLog( 'Error # ' + SFResultStr +
                  ', communications monitor file closed.' );
        Com.SniffMax := 0;
      End
    Else  { if file still ok, flush, check for wrap }
      If SniffSize > Com.SniffMax Then
        Begin
          Close( SniffLog );  SFResult := IOResult;
          SnifferOpen := False;
        End;
    {$I+}
    SniffBuffer := '';
  End;

{----------------------------------------------------------------------}
{ Used by sniffer to indicate change in port being accessed.           }
{----------------------------------------------------------------------}

Procedure CheckPortChange( PortNum:Byte );
  Begin
    If PortNum = PrevPort then Exit;
    SniffWrite( NL + '<Com' + Chr( PortNum + 48 ) + '> ' );
    PrevPort := PortNum;
  End;

{----------------------------------------------------------------------}
{ Returns true if requested serial port exists in the BIOS port table. }
{----------------------------------------------------------------------}

Function ComPortExists( PortNum:Byte ) : Boolean;
  Var
    ComPortTable :      { bios equipment table }
      Array [ 1..MaxComPort ] of Word absolute $40:0;
  Begin
    ComPortExists :=
       ( PortNum > 0 ) and ( PortNum <= MaxComPort ) and
       ( ComPortTable[ PortNum ] <> 0 );
  End;

{----------------------------------------------------------------------}
{ Toggles Data Terminal Ready line.                                    }
{----------------------------------------------------------------------}

Procedure DTROff( PortNum:Byte );
  Begin  Port[ ComPortAddr[ PortNum ] + ModemCtrlOfs ] := 0;  End;

{----------------------------------------------------------------------}

Procedure DTROn( PortNum:Byte );
  Begin  Port[ ComPortAddr[ PortNum ] + ModemCtrlOfs ] := 3;  End;

{----------------------------------------------------------------------}
{ Toggles Request To Send line.                                        }
{----------------------------------------------------------------------}

Procedure RTSOff( PortNum:Byte );
  Begin
    Delay( 1 );  Port[ ComPortAddr[ PortNum ] + ModemCtrlOfs ] := 1;
  End;

{----------------------------------------------------------------------}

Procedure RTSOn( PortNum:Byte );
  Begin  Port[ ComPortAddr[ PortNum ] + ModemCtrlOfs ] := 2;  End;

{----------------------------------------------------------------------}
{ Used by serial receive interrupt servers.                            }
{----------------------------------------------------------------------}

Procedure ComInServe( PortNum : Byte );
  Var IntrID : Byte;
  Begin
    Inline( $FA );  { disable interrupts }
    Repeat
      IntrID := Port[ ComPortAddr[ PortNum ] + IntrIDOfs ];
      Case IntrID shr 1 and $03 of
        0 : ;  { Modem status }
        1 : ;  { TX hold empty }
        2 : Begin  { RX data ready }
              ComRxBuffer[ PortNum ]^[ ComInPtr[ PortNum ] ] := 
                Port[ ComPortAddr[ PortNum ] ];
              ComInPtr[ PortNum ] := ( ComInPtr[ PortNum ] + 1 ) and $FF;
            End;
        3 : Begin
              ComRXErr[ PortNum ] := ComRXErr[ PortNum ] or  { receiver error }
                ( Port[ ComPortAddr[ PortNum ] + LineStatOfs ] and ComErrMask );
            End;
       End;
    Until IntrID and $01 <> 0;
    Port[ $20 ] := $20;  { clear intr }
    Inline( $FB );  { allow interrupts }
  End;

{----------------------------------------------------------------------}
{ Serial receive interrupt servers.                                    }
{----------------------------------------------------------------------}

{$F+}

Procedure ComIn1_ISR;  Interrupt;
  Begin  ComInServe( 1 );  End;

Procedure ComIn2_ISR;  Interrupt;
  Begin  ComInServe( 2 );  End;

{$F-}

{----------------------------------------------------------------------}
{ Setup serial port, returns false if port doesn't exist.              }
{----------------------------------------------------------------------}

Function ComInit( PortNum, Baud : Integer;  Parity : Char;
                  DataBits, StopBits : Integer;  ISR : Boolean ) : Integer;
  Var  Param, Junk : Byte;  Regs : Registers;
  Begin
    If Com.Fake[ PortNum ] Then
      Begin
        ComPortReady[ PortNum ] := True;
        ComInit := 0;  Exit;
      End;
    If not ComPortExists( PortNum ) Then
      ComResult := 1  { doesn't exist }
    Else
      If ComPortReady[ PortNum ] Then
        ComResult := 2  { already configured }
      Else
        ComResult := 0;  { OK }
    CheckPortChange( PortNum );
    If ComResult <> 0 Then
      SniffWrite( 'cannot init.' + NL )
    Else
      Begin   { port exist }
        ComISR[ PortNum ] := ISR;
        Case Baud of
          300 : Param := $40; { 010x xxxx }
         1200 : Param := $80; { 100x xxxx }
         2400 : Param := $A0; { 101x xxxx }
         4800 : Param := $C0; { 110x xxxx }
         Else   Param := $E0; { 111x xxxx };  { 9600 }
        End;
        If StopBits = 2 Then Param := Param or $04;   { xxxx x1xx }
        If DataBits = 7 Then
          Param := Param or $02  { xxxx xx10 }
        Else
          Param := Param or $03; { xxxx xx11 }
        If UpCase( Parity ) = 'O' Then
          Param := Param or $08  { xxxx 1xxx }
        Else
          If UpCase( Parity ) = 'E' Then
            Param := Param or $18;  { xxx1 1xxx }
        Regs.AH := $00;  { com init function }
        Regs.AL := Param;  Regs.DX := PortNum - 1;
        Intr( $14, Regs );
        If ComRxBuffer[ PortNum ] = NIL Then New( ComRxBuffer[ PortNum ] );
        ComInPtr[ PortNum ] := 0;  ComOutPtr[ PortNum ] := 0;
        Junk := Port[ ComPortAddr[ PortNum ] ]; { clear receiver }
        Junk := Port[ ComPortAddr[ PortNum ] + LineStatOfs ];
        Junk := Port[ ComPortAddr[ PortNum ] + ModemStatOfs ];
        { deactivate DTR ($01), RTS ($02), and OUT2 ($08) signals }
        Port[ ComPortAddr[PortNum] + ModemCtrlOfs ] := $0B;
        If ComISR[ PortNum ] Then
          Begin
         { enable interrupts for rx data ready ($01) and rx line stat ($04) }
            Port[ ComPortAddr[ PortNum ] + IntrEnableOfs ] := $05;
            Case PortNum of
              1 : Port[ $21 ] := Port[ $21 ] and $EF; { enable IRQ4 }
              2 : Port[ $21 ] := Port[ $21 ] and $F7; { enable IRQ3 }
             End;
            Case PortNum of
              1 : SetIntVec( $0C, @ComIn1_ISR );
              2 : SetIntVec( $0B, @ComIn2_ISR );
             End;
          End;
        SniffWrite( 'init ' + ComSetupStr( PortNum ) + NL );
        ComPortReady[ PortNum ] := True;
      End;  { port exists }
    ComInit := ComResult;
  End;

{----------------------------------------------------------------------}
{ True if transmit holding register is empty, ready for next byte.     }
{----------------------------------------------------------------------}

Function ComOutReady( PortNum : Byte ) : Boolean;
  Begin
    ComOutReady :=
      Port[ ComPortAddr[ PortNum ] + LineStatOfs ] and THRReadyMask <> 0;
  End;

{----------------------------------------------------------------------}
{ True if transmit shift register is empty.                            }
{----------------------------------------------------------------------}

Function ComOutDone( PortNum : Byte ) : Boolean;
  Begin
    ComOutDone :=
      Port[ ComPortAddr[ PortNum ] + LineStatOfs ] and TSRReadyMask <> 0;
  End;

{----------------------------------------------------------------------}
{ Waits for transmit to become completely empty.                       }
{----------------------------------------------------------------------}

Procedure ClearComOut( PortNum : Byte );
  Begin
    While ( not ComOutReady( PortNum ) ) or
          ( not ComOutDone( PortNum ) ) do;
    ComResult := 0;
  End;

{----------------------------------------------------------------------}
{ Clears a half duplex tx to prepare for rx.                           }
{----------------------------------------------------------------------}

Procedure ClearHalfDuplex( PortNum : Byte );
  Var Junk : Byte;
  Begin
    ClearComOut( PortNum );
    Junk := Port[ ComPortAddr[ PortNum ] ];
    Junk := Port[ ComPortAddr[ PortNum ] + LineStatOfs ];
    Junk := Port[ ComPortAddr[ PortNum ] + ModemStatOfs ];
    ComResult := 0;
  End;

{----------------------------------------------------------------------}
{ Output character to a serial port, not interrupt serviced.           }
{----------------------------------------------------------------------}

Procedure ComOut( PortNum, B : Byte );
  Var ComTimer : Word;  Junk : Byte;
  Begin
    ComResult := 0;
    If Com.Fake[ PortNum ] Then Exit;
    CheckPortChange( PortNum );
    SniffWrite( Chr( B ) );
    { wait for port to be ready to tx }
    ComTimer := Com.Timeout[ PortNum ];
    While ( not ComOutReady( PortNum ) ) and ( ComTimer > 0 ) do
      Dec( ComTimer );
    If ComTimer = 0 Then
      Begin  ComResult := TXCharTimeoutErr;  Exit;  End;
    { output byte }
    Port[ ComPortAddr[ PortNum ] ] := B;
  End;

{----------------------------------------------------------------------}
{ True if any characters are waiting in the receive buffer.            }
{----------------------------------------------------------------------}

Function ComInReady( PortNum : Byte ) : Boolean;
  Var Stat, Byt : Byte;
  Begin
    ComInReady :=
      ( ComISR[ PortNum ] and
        ( ( ComRXErr[ PortNum ] <> 0 ) or
          ( ComInPtr[ PortNum ] <> ComOutPtr[ PortNum ] ) ) ) or
      ( ( not ComISR[ PortNum ] ) and
        ( Port[ ComPortAddr[ PortNum ] + LineStatOfs ] and ComRXMask <> 0 ) );
  End;

{----------------------------------------------------------------------}
{ Clears the receive buffer.                                           }
{----------------------------------------------------------------------}

Procedure ClearComIn( PortNum : Byte );
  Var Junk : Byte;
  Begin
    While ComInReady( PortNum ) do
      Begin  Junk := ComIn( PortNum );  Delay( 40 );  End;
    ComResult := 0;
  End;

{----------------------------------------------------------------------}
{ Take a character from the interrupt serviced receive buffer.         }
{----------------------------------------------------------------------}

Function ComIn( PortNum : Byte ) : Byte;
  Var ComTimer : Word;  B, Junk : Byte;
  Begin
    ComResult := 0;  ComIn := 0;
    If Com.Fake[ PortNum ] Then Exit;
    CheckPortChange( PortNum );
    ComTimer := Com.Timeout[ PortNum ];
    While ( not ComInReady( PortNum ) ) and ( ComTimer > 0 ) do
      Dec( ComTimer );
    If ComTimer = 0 Then  { timeout error }
      Begin  ComResult := RXCharTimeoutErr;  Exit;  End;
    If not ComISR[ PortNum ] Then
      ComRXErr[ PortNum ] :=
        Port[ ComPortAddr[ PortNum ] + LineStatOfs ] and ComErrMask;
    If ComRXErr[ PortNum ] <> 0 Then
      Begin  { UART error }
        ComResult := ComPortError;
        If ComRXErr[ PortNum ] and $04 <> 0 Then
          ComResult := RxUARTParityErr;
        If ComRXErr[ PortNum ] and $02 <> 0 Then
          ComResult := RxUARTOverrunErr;
        If ComRXErr[ PortNum ] and $08 <> 0 Then
          ComResult := RxUARTFramingErr;
        If not ComISR[ PortNum ] Then
          Begin
            Junk := Port[ ComPortAddr[ PortNum ] ];
            Junk := Port[ ComPortAddr[ PortNum ] + LineStatOfs ];
            Junk := Port[ ComPortAddr[ PortNum ] + ModemStatOfs ];
          End;
        ComRXErr[ PortNum ] := 0;  Exit;
      End;  { UART error }
    If ComISR[ PortNum ] Then
      Begin  { get char }
        B := ComRxBuffer[ PortNum ]^[ ComOutPtr[ PortNum ] ];
        ComOutPtr[ PortNum ] := ( ComOutPtr[ PortNum ] + 1 ) and $FF;
      End
    Else
      B := Port[ ComPortAddr[ PortNum ] ];
    SniffWrite( Chr( B ) );
    ComIn := B;
  End;

{----------------------------------------------------------------------}
{ Verifies external connection to serial port by checking for echo.    }
{----------------------------------------------------------------------}

Function LineOK( PortNum, EchoByte : Byte ) : Boolean;
  Begin
    ComOut( PortNum, EchoByte );
    LineOK := ( ComIn( PortNum ) = EchoByte ) and ( ComResult = 0 );
  End;

{----------------------------------------------------------------------}
{ Returns a serial port setup by interrogating the hardware.           }
{----------------------------------------------------------------------}

Procedure ComInq( PortNum : Integer;  Var Baud : Integer;
                  Var Parity : Char;  Var DataBits, StopBits : Integer;
                  Var Duplex : Char;  Var ISR : Boolean );
  Var  Temp : Word;
  Begin
    Baud := 0;  StopBits := 0;  DataBits := 0;  Parity := 'X';
    Duplex := 'X';  ISR := False;
    If not ComPortExists( PortNum ) Then Exit;
    ISR := ComISR[ PortNum ];
    Port[ ComPortAddr[ PortNum ] + LineCtrlOfs ] :=
      Port[ ComPortAddr[ PortNum ] + LineCtrlOfs ] or $80;
    Temp := ( Port[ ComPortAddr[ PortNum ] + BaudDivOfs ]  shl 8 ) +
            Port[ ComPortAddr[ PortNum ] ];
    Port[ ComPortAddr[ PortNum ] + LineCtrlOfs ] :=
      Port[ ComPortAddr[ PortNum ] + LineCtrlOfs ] and $7F;
    Case Temp of
      $0006 : Baud := 19200;
      $000C : Baud := 9600;
      $0018 : Baud := 4800;
      $0030 : Baud := 2400;
      $0060 : Baud := 1200;
      $00C0 : Baud := 600;
      $0180 : Baud := 300;
      $0300 : Baud := 150;
      Else Baud := 0;
     End;
    Temp := Port[ ComPortAddr[ PortNum ] + LineCtrlOfs ] and $0003;
    Case Temp of
      0 : DataBits := 5;
      1 : DataBits := 6;
      2 : DataBits := 7;
      3 : DataBits := 8;
      Else DataBits := 0;
     End;
    Temp := Port[ ComPortAddr[ PortNum ] + LineCtrlOfs ] and $0004;
    If Temp = 0 THEN StopBits := 1 ELSE StopBits := 2;
    Temp := Port[ ComPortAddr[ PortNum ] + LineCtrlOfs ] and $0038;
    Case Temp OF
      $18 : Parity := 'E';
      $08 : Parity := 'O';
      $28 : Parity := 'S';
      $00 : Parity := 'N';
      Else Parity := 'X';
     End;
  End;

{-------------------------------------------------------------------------}
{ Returns a serial port setup as a string bbbbbpdsdi (ie. 9600N81FI).     }
{ where bbbbb=baud 18200..150, p=parity O/E/N, d=data bits 5..8,          }
{ s=stop bits 1..2, d=duplex H/F, i=ISR I/-                               }
{-------------------------------------------------------------------------}

Function ComSetupStr( ComPort : Byte ) : ComSetupStrType;
  Var Baud, StopBits, DataBits : Integer;
      Parity, Duplex : Char;
      StrTemp, OutTemp : ComSetupStrType;
      ISR : Boolean;
  Begin
    ComInq( ComPort, Baud, Parity, DataBits, StopBits, Duplex, ISR );
    Str( Baud, StrTemp );  OutTemp := StrTemp + Parity;
    Str( DataBits, StrTemp );  OutTemp := OutTemp + StrTemp;
    Str( StopBits, StrTemp );  OutTemp := OutTemp + StrTemp;
    OutTemp := OutTemp + Duplex;
    If ISR Then OutTemp := OutTemp + 'I' Else OutTemp := OutTemp + '-';
    ComSetupStr := OutTemp;
  End;

{-------------------------------------------------------------------------}
{ Returns serial port status info as a string where capital letters       }
{ indicate logical one state.  String is 'drooldopfbtt' representing      }
{ Modem status:                                                           }
{   [C]lear to send, [D]ata set ready, [R]ing indicator,                  }
{   [R]eceive line signal detect.                                         }
{ Modem control:                                                          }
{   [D]ata terminal ready, [R]equest to send, [O]ut 1, [O]ut 2, [L]oop.   }
{ Line status:                                                            }
{   [D]ata ready, [O]verrun error, [P]arity error, [F]raming error,       }
{   [B]reak error, [T]x hold empty, [T]x shift empty.                     }
{-------------------------------------------------------------------------}

Function ComStatStr( PortNum : Byte ) : ComStatStrType;
  Var  ModmStat, ModmCtrl, LineStat : Byte;  O : ComStatStrType;
  Begin
    ModmStat := Port[ ComPortAddr[ PortNum ] + ModemStatOfs ];
    ModmCtrl := Port[ ComPortAddr[ PortNum ] + ModemCtrlOfs ];
    LineStat := Port[ ComPortAddr[ PortNum ] + LineStatOfs ];
    O := 'cdrrdrooldopfbtt';
    If ModmStat and $10 <> 0 Then O[ 1 ] := UpCase( O[ 1 ] );
    If ModmStat and $20 <> 0 Then O[ 2 ] := UpCase( O[ 2 ] );
    If ModmStat and $40 <> 0 Then O[ 3 ] := UpCase( O[ 3 ] );
    If ModmStat and $80 <> 0 Then O[ 4 ] := UpCase( O[ 4 ] );
    If ModmCtrl and $01 <> 0 Then O[ 5 ] := UpCase( O[ 5 ] );
    If ModmCtrl and $02 <> 0 Then O[ 6 ] := UpCase( O[ 6 ] );
    If ModmCtrl and $04 <> 0 Then O[ 7 ] := UpCase( O[ 7 ] );
    If ModmCtrl and $08 <> 0 Then O[ 8 ] := UpCase( O[ 8 ] );
    If ModmCtrl and $10 <> 0 Then O[ 9 ] := UpCase( O[ 9 ] );
    If LineStat and $01 <> 0 Then O[ 10 ] := UpCase( O[ 10 ] );
    If LineStat and $02 <> 0 Then O[ 11 ] := UpCase( O[ 11 ] );
    If LineStat and $04 <> 0 Then O[ 12 ] := UpCase( O[ 12 ] );
    If LineStat and $08 <> 0 Then O[ 13 ] := UpCase( O[ 13 ] );
    If LineStat and $10 <> 0 Then O[ 14 ] := UpCase( O[ 14 ] );
    If LineStat and $20 <> 0 Then O[ 15 ] := UpCase( O[ 15 ] );
    If LineStat and $40 <> 0 Then O[ 16 ] := UpCase( O[ 16 ] );
    ComStatStr := O;
  End;

{----------------------------------------------------------------------}
{ Deactivates interrupt service before program terminates.             }
{----------------------------------------------------------------------}

{$F+}
Procedure ComUnload;
  Var Junk : Integer;
  Begin
    If Com.SniffMax > 0 Then
      Begin
        SniffFlush;  SnifferOpen := False;
        {$I-} Close( SniffLog ); {$I+} Junk := IOResult;
      End;
    ExitProc := ComExitSave;
    If ComRxBuffer[ 1 ] <> NIL Then
      Port[ $21 ] := Port[ $21 ] or $10;  { IRQ4 disable }
    If ComRxBuffer[ 2 ] <> NIL Then
      Port[ $21 ] := Port[ $21 ] or $08;  { IRQ3 disable }
  End;
{$F-}

{----------------------------------------------------------------------}
{ Call InitComSystem once at program startup.                          }
{----------------------------------------------------------------------}

Procedure InitComSystem;
  Begin
    ComExitSave := ExitProc;
    ExitProc := @ComUnload;
    FillChar( Com, SizeOf( Com ), 0 );
    For PrevPort := 1 to MaxComPort do
      Begin
        ComRxBuffer[ PrevPort ] := NIL;
        ComPortReady[ PrevPort ] := False;
        ComRXErr[ PrevPort ] := 0;
        Com.Timeout[ PrevPort ] := ComDefaultTimeout;
        ComISR[ PrevPort ] := False;
      End;
    ComResult := 0;
    SniffBuffer := '';
    SniffSize := 0;
    SniffBufferMax := 0;
    ErrorCapture := False;
    SniffLineLength := 0;
    PrevPort := 100;
    SniffOverFlow := False;
    SnifferOpen := False;
  End;

{----------------------------------------------------------------------}

Begin

  InitComSystem;

End.
