(*********************************************************************
**********************************************************************
****                                                              ****
****                       DISK.PAS                               ****
****                                                              ****
****          Disk utilities, logs, and error handling.           ****
****                                                              ****
**********************************************************************
**********************************************************************)


Unit Disk;

{$I LIBRARY.CFG }

INTERFACE

Uses Dos, Crt, Global, Sounds, Math, Strings, Video, Time, Key;

CONST
  MaxLogFileSize : LongInt = 1000000;  { 1 meg }
  ChkErrRecursive : Boolean = False;
  ProgramLogRecursive : Boolean = False;
  ErrorLogRecursive : Boolean = False;
  AuditLogRecursive : Boolean = False;
  AuditLogOpen : Boolean = False;
  _FileNotFound = 2;
  _SeekError = 156;


VAR
  Result, CloseResult, JunkResult : Integer;


FUNCTION  FileExists( ChkName:String40 ):Boolean;
FUNCTION  DiskMounted( Drive:Char ):Boolean;
FUNCTION  SizeOfFile( TextFileName:GenlString ):LongInt;
FUNCTION  WriteTextFile( TextFileName:String40; TextIn:String ):Integer;
PROCEDURE AuditLog( Msg:String100 );
PROCEDURE ProgramLog( Msg:String100 );
PROCEDURE ErrorLog( Msg:String100 );
FUNCTION  ErrorName( Err : Integer ):String30;
Procedure Stop( Msg : GenlString );
Procedure Quit( Msg : GenlString );
FUNCTION  CopyFile( Name1, Name2 : GenlString ) : Integer;
FUNCTION  RenameFile( Name1, Name2 : GenlString ) : Integer;
FUNCTION  EraseFile( Name : GenlString ) : Integer;
FUNCTION  NewExt( Name : GenlString;  Ext : String5 ) : GenlString;


IMPLEMENTATION

{--------------------------------------------------------------------------}
{ FUNCTION FileExists                                                      }
{                                                                          }
{ Returns true if the file exists in the current directory and false if it }
{ does not.                                                                }
{--------------------------------------------------------------------------}

FUNCTION FileExists( ChkName:String40 ):Boolean;
  VAR ChkResult:Integer; ChkFile:File;
  BEGIN
    Assign( ChkFile, ChkName);
    {$I-} Reset( ChkFile ); {$I+} ChkResult := IOResult;
    FileExists := ChkResult = 0;
    {$I-} Close( ChkFile ); {$I+} ChkResult := IOResult;
  END;


{--------------------------------------------------------------------------}
{ FUNCTION DiskMounted                                                     }
{                                                                          }
{ Returns true if the drive specified by the parameter exists.             }
{--------------------------------------------------------------------------}

FUNCTION DiskMounted( Drive:Char ):Boolean;
  VAR
    Regs : Registers;
  BEGIN
    If Upcase( Drive ) = 'B'    { TRK 0, DRV 0/1 }
      THEN Regs.DX := 1 ELSE Regs.DX := 0; { default to A }
    Regs.AX := $0401;  { send verify 1 sect }
    Regs.CX := 1;      { sect 1 }
    Intr( $13, Regs );
    DiskMounted := Regs.AX and $8000 = 0;
  END;

{--------------------------------------------------------------------------}
{ FUNCTION SizeOfFile                                                      }
{                                                                          }
{ Returns the size in characters if the file.                              }
{--------------------------------------------------------------------------}

FUNCTION SizeOfFile( TextFileName:GenlString ):LongInt;

VAR
  CharFile:File of Char;
  Size:LongInt;
  SizeResult:Integer;

BEGIN
  Assign( CharFile, TextFileName );
  {$I-} Reset( CharFile ); {$I+} SizeResult := IOResult;
  If SizeResult = 0
    THEN SizeOfFile := FileSize( CharFile )
    ELSE SizeOfFile := 0;
  {$I-} Close( CharFile ); {$I+} SizeResult := IOResult;
END;

{--------------------------------------------------------------------------}
{ FUNCTION WriteTextFile                                                   }
{                                                                          }
{ Writes a string to a textfile and returns the IOResult of the operation. }
{--------------------------------------------------------------------------}

FUNCTION WriteTextFile( TextFileName:String40; TextIn:String ):Integer;
  VAR  TextLog:Text;  CloseResult, LogResult:Integer;
  BEGIN
    If SizeOfFile( TextFileName ) > MaxLogFileSize
      THEN LogResult := 101 ELSE LogResult := 0;
    If LogResult = 0
      THEN BEGIN
        Assign( TextLog, TextFileName );
        {$I-} Append( TextLog ); {$I+} LogResult := IOResult;
        If LogResult <> 0
          THEN BEGIN    { cannot append }
            {$I-} ReWrite( TextLog ); {$I+} LogResult := IOResult;
          END;     { cannot append }
      END;
    If LogResult = 0
      THEN BEGIN    { file ok }
        {$I-} Writeln( TextLog, TextIn ); {$I+} LogResult := IOResult;
      END;
    {$I-} Close( TextLog ); {$I+} CloseResult := IOResult;
    If LogResult = 0 then LogResult := CloseResult;
    WriteTextFile := LogResult;
  END;

{--------------------------------------------------------------------------}
{ PROCEDURE AuditLog                                                       }
{                                                                          }
{ Writes a message to the Audit Log File.                                  }
{--------------------------------------------------------------------------}

PROCEDURE AuditLog( Msg:String100 );
  VAR  AuditLogResult : Integer;
       LogSpec : GenlString;
  BEGIN
    If Sys.Silent or AuditLogRecursive or
      ( not AuditLogOpen ) THEN Exit;
    AuditLogRecursive := True;
    LogSpec := Date6( DateString );
    Delete( LogSpec, 5, 1 );  { delete year decade digit }
    LogSpec := Sys.LogPath + LogSpec + Sys.InstallCode + '.LOG';
    AuditLogResult :=
      WriteTextFile( LogSpec,
                     DateString + ' ' + TimeString + ' ' + Msg );
    ChkErr( AuditLogResult, 'W', LogSpec, 'audit log' );
    AuditLogOpen := AuditLogResult = 0;
    AuditLogRecursive := False;
  END;

{--------------------------------------------------------------------------}
{ PROCEDURE ProgramLog                                                     }
{                                                                          }
{ Writes a message to the Program Log File.                                }
{--------------------------------------------------------------------------}

PROCEDURE ProgramLog( Msg:String100 );

VAR
  ProgramLogResult : Integer;

BEGIN
{ entries also go to audit log }
  If Sys.Silent or ProgramLogRecursive THEN Exit;
  ProgramLogRecursive := True;
  AuditLog( Msg );
  ProgramLogResult :=
    WriteTextFile( '\RUN' + Sys.InstallCode + '.LOG',
                   DateString + ' ' + TimeString + ' ' + Msg );
  ChkErr( ProgramLogResult, 'W',
          '\RUN' + Sys.InstallCode + '.LOG', 'program log' );
  ProgramLogRecursive := False;
END;

{--------------------------------------------------------------------------}
{ PROCEDURE ErrorLog                                                       }
{                                                                          }
{ Writes a message to the Error Log File.                                  }
{--------------------------------------------------------------------------}

PROCEDURE ErrorLog( Msg:String100 );
  VAR  ErrorLogResult : Integer;
  BEGIN
  { entries also go to program log, and from there to audit log }
    If Sys.Silent or ErrorLogRecursive THEN Exit;
    ErrorLogRecursive := True;
    ProgramLog( Msg );
    ErrorLogResult :=
      WriteTextFile( '\ERROR' + Sys.InstallCode + '.LOG',
                     DateString + ' ' + TimeString + ' ' + Msg );
    ChkErr( ErrorLogResult, 'W',
            '\ERROR' + Sys.InstallCode + '.LOG', 'error log' );
    ErrorLogRecursive := False;
  END;

{--------------------------------------------------------------------------}
{ FUNCTION ErrorName                                                       }
{                                                                          }
{ Given an error number, this returns the name of that error.              }
{--------------------------------------------------------------------------}

FUNCTION ErrorName( Err : Integer ):String30;
  BEGIN
    ErrorName := 'Error # ' + I_S( Err );
    Case Err of
         0 : ErrorName := 'No error';
         2 : ErrorName := 'Error, file not found';
         3 : ErrorName := 'Error, path not found';
         4 : ErrorName := 'Error, too many files open';
         5 : ErrorName := 'Error, access denied';
        17 : ErrorName := 'Error, inter-device rename';
       100 : ErrorName := 'Error, access past file end';
       101 : ErrorName := 'Error, disk full';
       102 : ErrorName := 'Error, file not assigned';
       103 : ErrorName := 'Error, file not open';
       104 : ErrorName := 'Error, not open for input';
       105 : ErrorName := 'Error, not open for output';
       106 : ErrorName := 'Bad numeric format';
       150 : ErrorName := 'Error, write protected';
       152 : ErrorName := 'Error, drive not ready';
       154 : ErrorName := 'CRC error in data';
       156 : ErrorName := 'Seek error';
       157 : ErrorName := 'Error, unknown disk type';
       158 : ErrorName := 'Error, sector not found';
       159 : ErrorName := 'Printer out of paper';
       160 : ErrorName := 'Printer off line';
       161 : ErrorName := 'Device read error';
       162 : ErrorName := 'Hardware failure';
       200 : ErrorName := 'Error, division by zero';
       201 : ErrorName := 'Range check error';
       202 : ErrorName := 'Stack overflow error';
       203 : ErrorName := 'Heap overflow error';
       255 : ErrorName := 'User break';
       -94 : ErrorName := 'Insufficient memory';
       -95 : ErrorName := 'Too many data handles';
       -96 : ErrorName := 'Bad manager function';
       -97 : ErrorName := 'Initialization error';
       -98 : ErrorName := 'Record size error';
       -99 : ErrorName := 'Process error';
      -100 : ErrorName := 'Pointer size underflow error';
      -101 : ErrorName := 'Pointer size overflow error';
      -102 : ErrorName := 'Pointer underflow error';
      -103 : ErrorName := 'Pointer overflow error';
      -110 : ErrorName := 'Screen size underflow error';
      -111 : ErrorName := 'Screen size overflow error';
      -112 : ErrorName := 'Screen underflow error';
      -113 : ErrorName := 'Screen overflow error';
     End;
  END;

{--------------------------------------------------------------------------}
{ PROCEDURE FullChkErr                                                     }
{                                                                          }
{ Replaces gobal ChkErr for use throughout program.                        }
{--------------------------------------------------------------------------}

PROCEDURE FullChkErr( Result:integer; Errtype:Char;
                      FSpec:GenlString; Desc:GenlString );
  VAR  Msg:String40;  UsrMsg, LogMsg:GenlString;
  BEGIN
    If ChkErrRecursive THEN Exit;
    If Result <> 0
      THEN BEGIN
        ChkErrRecursive := True;
        If FSpec = '' THEN FSpec := Desc;
        If Desc = '' THEN Desc := FSpec;
        If Result = _FileNotFound
          THEN BEGIN
            Msg := ' file not found';
            UsrMsg := Desc + Msg + '.';
            LogMsg := FSpec + Msg + '.';
            UsrMsg[ 1 ] := UpCase( UsrMsg[ 1 ] );
            LogMsg[ 1 ] := UpCase( LogMsg[ 1 ] );
          END
          ELSE BEGIN
            Msg := ErrorName( Result );
            Case Errtype of
              'R', 'r' : Msg := Msg + ' reading ';
              'W', 'w' : Msg := Msg + ' writing ';
              'O', 'o' : Msg := Msg + ' opening ';
              'C', 'c' : Msg := Msg + ' closing ';
              'S', 's' : Msg := Msg + ' seeking ';
              'U', 'u' : Msg := Msg + ' using ';
              Else Msg := Msg + ' ';
             End; { case }
            UsrMsg := Msg + Desc + '.';
            LogMsg := Msg + FSpec + '.';
          END;
        AuditLog( 'Error in: ' + FullOperation );
        If UpCase( ErrType ) <> ErrType
          THEN BEGIN
            ErrorLog( LogMsg + ' - FATAL' );
            Message( UsrMsg, Msg_SirenCallHalt );
            Halt( 253 );  { pass error on to TidyExit }
          END
          ELSE BEGIN
            AuditLog( LogMsg );  Message( UsrMsg, Msg_OopKey );
          END;
        ChkErrRecursive := False;
      END;
  END;

{--------------------------------------------------------------------------}
{ PROCEDURE Stop                                                           }
{                                                                          }
{ Present a message and halt program with alert.                           }
{--------------------------------------------------------------------------}

Procedure Stop( Msg : GenlString );
  Begin
    ErrorLog( Msg + ' - STOPPED' );
    Message( Msg, Msg_SirenCallHalt );
    Halt( 1 );  { pass error on to TidyExit }
  End;

{--------------------------------------------------------------------------}
{ PROCEDURE Quit                                                           }
{                                                                          }
{ Present a message and halt program calmly.                               }
{--------------------------------------------------------------------------}

Procedure Quit( Msg : GenlString );
  Begin
    ErrorLog( Msg + ' - QUIT' );
    Message( Msg, Msg_OopKeyHalt );
    Halt;
  End;

{--------------------------------------------------------------------------}
{ FUNCTION CopyFile                                                        }
{                                                                          }
{ Copies the file with Name1 to a new file with Name2 and returns the      }
{ IOResult of the operation.                                               }
{--------------------------------------------------------------------------}

FUNCTION CopyFile( Name1, Name2 : GenlString ) : Integer;

LABEL CopyFileOut;

VAR  InFile, OutFile : File;
     NumRead, TotalBlocks, CurrentBlock : Word;
     CopyResult : Integer;
     Buffer : Array [ 1..1024 ] of Byte;

BEGIN
  {$I-}
  Assign( InFile, Name1 );
  Assign( OutFile, Name2 );
  Reset( InFile, 1 );  CopyResult := IOResult;
  If CopyResult <> OK THEN Goto CopyFileOut;
  TotalBlocks := FileSize( InFile );
  ReWrite( OutFile, 1 );  CopyResult := IOResult;
  If CopyResult <> OK THEN Goto CopyFileOut;
  Disp( 'Copying file...  ' );  ShowPercent( 0.0 );
  TotalBlocks := FileSize( InFile ) div SizeOf( Buffer ) + 1;
  CurrentBlock := 0;
  While ( CopyResult = OK ) and ( not EOF( InFile ) ) do
    BEGIN
      BlockRead( InFile, Buffer, SizeOf( Buffer ), NumRead );
      CopyResult := IOResult;
      If CopyResult = OK
        THEN BEGIN
          BlockWrite( OutFile, Buffer, NumRead );
          CopyResult := IOResult;
        END;
      Inc( CurrentBlock );
      ShowPercent( CurrentBlock / TotalBlocks );
    END;
CopyFileOut:
  Close( InFile );  CloseResult := IOResult;
  If CopyResult = OK THEN CopyResult := CloseResult;
  Close( OutFile );  CloseResult := IOResult;
  If CopyResult = OK THEN CopyResult := CloseResult;
  CopyFile := CopyResult;
  {$I+}
END;

{--------------------------------------------------------------------------}
{ FUNCTION RenameFile                                                      }
{                                                                          }
{ Renames the file with Name1 to Name2 and returns the IOResult of the     }
{ operation.                                                               }
{--------------------------------------------------------------------------}

FUNCTION RenameFile( Name1, Name2 : GenlString ) : Integer;

VAR  RFile : File;  RResult : Integer;

BEGIN
  Assign( RFile, Name1 );
  {$I-} ReName( RFile, Name2 ); {$I+} RResult := IOResult;
  RenameFile := RResult;
END;

{--------------------------------------------------------------------------}
{ FUNCTION EraseFile                                                       }
{                                                                          }
{ Erases the file whose name is passed to the function and returns the     }
{ IOResult of the operation.                                               }
{--------------------------------------------------------------------------}

FUNCTION EraseFile( Name : GenlString ) : Integer;
  VAR  EFile : File;  EResult : Integer;
  BEGIN
    Assign( EFile, Name );
    {$I-} Erase( EFile ); {$I+} EResult := IOResult;
    If EResult = _FileNotFound THEN EResult := OK;
    EraseFile := EResult;
  END;

{--------------------------------------------------------------------------}
{ FUNCTION NewExt                                                          }
{                                                                          }
{ Returns file name with extension added or replaced.                      }
{--------------------------------------------------------------------------}

FUNCTION NewExt( Name : GenlString;  Ext : String5 ) : GenlString;
  VAR  ExtPos : Byte;
  BEGIN
    NewExt := Name;  If Name = '' THEN Exit;
    ExtPos := Pos( '.', Name );  If ExtPos = 1 THEN Exit;
    If ExtPos = 0 THEN ExtPos := Length( Name ) + 1;
    NewExt := Copy( Name, 1, ExtPos - 1 ) + '.' + Ext;
  END;

{==========================================================================}

BEGIN

  ChkErr := FullChkErr;

END.

