(*********************************************************************
**********************************************************************
****                                                              ****
****                        INIT.PAS                              ****
****                                                              ****
****           Initialization and shutdown procedures.            ****
****                                                              ****
**********************************************************************
**********************************************************************)


(* This initialization unit contains important utilities for
   loading a configuration file, checking command line options,
   logging program startup and end, providing basic command line help,
   and handling fatal errors.  Basic usage requires one or two lines
   at the very beginning of the main program block:
     ProgramHelp := CommandLineHelp;  { optional }
     Initialize( ProgramName, Version, Revision );
   "ProgramHelp" sets the procedure which will be executed if a  "?"  is
     used on the command line.  Generally, a basic program description
     and program specific command line options.  This line can be ommitted
     if the program provides no description or program specific help.
   "Initialize" sets the global program name, version, revision information.
     checks for command line help requests,
     reads the command line options, reads the initialization file,
     and logs program startup.
   The TidyExit procedure address is automatically set at unit startup. *)


Unit Init;


{$I LIBRARY.CFG }


INTERFACE


Uses Dos, Crt, Global, Sounds, Math, Strings, Video, Time, Key, Disk;


VAR
  ExitSave : Pointer;
  ProgramHelp : Procedure;
  CheckInitLine : Function( InitLine : GenlString ) : Boolean;

CONST
  { Ini statements are separated here so they can be accessed by parent
    program, such as a preference manager, to detect unknown statements. }
  MaxIniWords = 26;
  OutDev : GenlString = 'CON';
  IniWords : Array[ 1..MaxIniWords ] of String[10] =
     ( ';', 'PLA', 'SOI', 'MIS', 'CLE', 'TAN', 'LOG', 'INS',
       'HEL', 'SYSTEMDATE', 'SOU', 'COM', 'YEA', 'SYSTEMTYPE',
       'TER', 'DATEF', 'SIL', 'KEYLOG', 'SCR', 'DISK', 'OUTDEV',
       'BEGIN', 'END', 'MAS', 'DATA', 'EORUN' );
  TidyCode : Integer = 0;

PROCEDURE InitLog( Msg:String40 );
PROCEDURE Initialize( NameIn, VerIn, RevIn : String40 );


IMPLEMENTATION

{--------------------------------------------------------------------------}
{ PROCEDURE InitLog                                                        }
{                                                                          }
{ Called by ReadInit, this procedure writes a message to the screen if     }
{ Sys.DebugEnabled is set to true.                                         }
{--------------------------------------------------------------------------}

PROCEDURE InitLog( Msg:String40 );

BEGIN
  If Sys.DebugEnabled THEN DispLn( Msg );
END;


{--------------------------------------------------------------------------}
{ PROCEDURE DefaultProgramHelp                                             }
{                                                                          }
{ This procedure may be replaced in the main program block by assigning to }
{ a user written procedure.                                                }
{--------------------------------------------------------------------------}

PROCEDURE DefaultProgramHelp;

BEGIN
  DispLn( 'No program specific help is available.' );
END;


{--------------------------------------------------------------------------}
{ PROCEDURE ShowProgramHelp                                                }
{                                                                          }
{ Writes a message to the Audit Log File that Help has been requested and  }
{ calls the help routines, if they exist for that program.                 }
{--------------------------------------------------------------------------}

PROCEDURE ShowProgramHelp;
  VAR TempVector : Word;
  BEGIN
    AuditLog( 'Command line program help requested.' );
    DispLn( NL + 'Program ' + FullName + '.' );
    If Sys.Demo
      THEN DispLn( 'This demonstration version does not provide help.' )
      ELSE BEGIN
        ProgramHelp;
        DispLn( 'Enter  ' + Sys.ProgramName +
                 ' ??  for standard options.' + NL );
      END;
    Halt;
  END;


{--------------------------------------------------------------------------}
{ PROCEDURE ShowStandardHelp                                               }
{                                                                          }
{ Shows the standard command line options.                                 }
{--------------------------------------------------------------------------}

PROCEDURE ShowStandardHelp;

BEGIN
  AuditLog( 'Command line options help requested.' );
  DispLn( NL + 'Program ' + FullName + '.' );
  If Sys.Demo
    THEN DispLn( 'This demonstration version does not provide help.' )
    ELSE BEGIN
      DispLn( 'System wide options:' );
      DispLn( '  /D  (debug) activates diagnostic displays.' );
      DispLn( '  /D2  (debug2) force install code 000.' );
      DispLn( '  /F  (future) run one week in advance of work dates.' );
      DispLn( '  /I fn  (init) use alternate config file.' );
      DispLn( '  /M  (memory) skip RAM and disk checks.' );
      DispLn( '  /P  (printer) don''t check for printer.' );
      DispLn( '  /S  (sound) deactivates extended sounds.' );
      DispLn( '  /W  (week) use system date for week start work dates.' );
      DispLn( '  SHARE  in environment allows shared local/remote use.' );
      DispLn( '  REMOTE  in environment simulates remote use.' );
      DispLn( '  ERRORLEVEL returns:  0=OK, 1=app returns,' );
      DispLn( '    2..252=crash, 253=fatal, 254=timeout, 255=break.' );
      DispLn( 'Enter  ' + Sys.ProgramName +
               ' ???  for configuration help.' );
      DispLn( '' );
    END;
  Halt;
END;

{--------------------------------------------------------------------------}
{ PROCEDURE ShowConfigHelp                                                 }
{                                                                          }
{ Shows the standard configuration file statements.                        }
{--------------------------------------------------------------------------}

PROCEDURE ShowConfigHelp;

BEGIN
  AuditLog( 'Command line configuration help requested.' );
  DispLn( NL + 'Program ' + FullName + '.' );
  If Sys.Demo
    THEN DispLn( 'This demonstration version does not provide help.' )
    ELSE BEGIN
      DispLn( 'Configuration file statements:' );
      DispLn( '  ; ...   (comment)' );
      DispLn( '  CLEanpath ...  (ship/bill data path)' );
      DispLn( '  COMpath ...  (communications data path)' );
      DispLn( '  DATApath ...  (data file path)' );
      DispLn( '  DATEFormat n  (0..2=mmddyy,ddmmyy, yymmdd)' );
      DispLn( '  EURONumbers on  (swap decimal/comma in number show)' );
      DispLn( '  HELppath ...  (help file path)' );
      DispLn( '  INStallation ppt  (plant #pp, terminal #t)' );
      DispLn( '  KEYboardlogging on  (record all keystrokes)' );
      DispLn( '  LOGpath ...  (audit log path)' );
      DispLn( '  MASterpath ...  (master file path)' );
      DispLn( '  MIScpath ...  (printer config path)' );
      DispLn( '  OUTDEV ...  (name of printer device)' );
      DispLn( '  PLAntname ...  (installation name)' );
      DispLn( '  SCReensavertime seconds  (allowed inactive seconds)' );
      DispLn( '    (maximum keypress wait, < 0 = quit after timeout)' );
      DispLn( '  SILent on  (all logs inhibited)' );
      DispLn( '  SOIlpath ...  (soil file path)' );
      Disp( 'PRESS ANY KEY . . .' + #13 );  AttentionSound;  WaitKeyPress;
      DispLn( '  SOUnd off  (deactivate extended sounds)' );
      DispLn( '  SYSTEMDate on  (use system date for week start)' );
      DispLn( '  SYSTEMType ... (software type for screen heading)' );
      DispLn( '  TERminalName ... (name of computer)' );
      DispLn( '  TANkenable on  (enable tank features)' );
      DispLn( '  YEArrange minyear maxyear (specify allowed range)' );
      DispLn( 'Individual programs may have ' +
              'specific additional statements.' );
      DispLn( 'Enter  ' + Sys.ProgramName +
               ' ?  for program specific help.' );
      DispLn( '' );
    END;
  Halt;
END;


{--------------------------------------------------------------------------}
{ FUNCTION DiskFreeCheck                                                   }
{                                                                          }
{ Calls DiskFree to get the # of bytes free on the disk and displays a     }
{ warning message if the diskspace is low or empty.                        }
{--------------------------------------------------------------------------}

Function DiskFreeCheck( Arg : GenlString ) : Integer;

VAR  ActualFree, WarnFree, MinFree:LongInt;  DriveNum:Byte;

BEGIN
  DiskFreeCheck := -99;
  DriveNum := ParseInt( Arg );
  If ConversionError THEN Exit;
  WarnFree := ParseInt( Arg );
  If ConversionError THEN Exit;
  MinFree := ParseInt( Arg );
  If ConversionError THEN Exit;
  InitLog( 'Disk ' + Chr( 64 + DriveNum ) + ' check: ' +
           I_S( WarnFree ) + ', ' + I_S( MinFree ) );
  ActualFree := DiskFree( DriveNum );
  If ActualFree < MinFree
    THEN ChkErr( -99, 'x', 'Out of disk space.', '' );
  If ActualFree < WarnFree
    THEN ChkErr( -99, 'X', 'Disk space low.', '' );
  DiskFreeCheck := 0;
END;


{--------------------------------------------------------------------------}
{ PROCEDURE ReadInit                                                       }
{                                                                          }
{ Called by Initialize, this procedure reads the command line and the      }
{ configuration file and sets the system up accordingly.                   }
{--------------------------------------------------------------------------}

PROCEDURE ReadInit;

VAR  Parse, IniCode, IniWord, IniLoop: Integer;
     IniFile:Text;
     RawIniLine, IniLine, State, Arg: GenlString;
     ValidStuff : Boolean;

CONST  IniFileSpec : String40 = '\SYSTEM.INI';

{---}
PROCEDURE IniPath( Msg:String40; VAR InPath:String40 );

  BEGIN
    InPath := Arg;
    If ( InPath <> '' ) and ( InPath[ Length( InPath ) ] <> '\' )
      THEN InPath := InPath + '\';
    InitLog( Msg + ' path: ' + InPath );
  END;
{---}

BEGIN
{ read command line }
  If ParamScan( '?' ) > 0 THEN ShowProgramHelp;
  If ParamScan( '??' ) > 0 THEN ShowStandardHelp;
  If ParamScan( '???' ) > 0 THEN ShowConfigHelp;
  If FullYear < 1989 THEN ChkErr( -99, 'x', 'System clock failed.', '' );
  Sys.UseFutureDate := ParamScan( '/F' ) > 0;
  Parse := ParamScan( '/I' );
  If Parse > 0
    THEN BEGIN
      If Parse >= ParamCount
        THEN ChkErr( 99, 'x', '', 'Initialization file name missing' );
      IniFileSpec := ParamStr( Parse + 1 );
      IniFileSpec := Upper( IniFileSpec );
    END;
  Sys.PrinterAbandoned := ParamScan( '/P' ) > 0;
  Sys.SoundOn := ParamScan( '/S' ) = 0;
  Sys.UseSystemDate := ParamScan( '/W' ) > 0;
  Sys.IgnoreLowMemory := ParamScan( '/M' ) > 0;
{ read configuration file }
  ValidStuff := True;
  Assign( IniFile, IniFileSpec );
  {$I-} Reset( IniFile ); {$I+} IniCode := IOResult;
  While ( IniCode = 0 ) and ( not EOF(IniFile) ) Do
    BEGIN  { file open and not eof }
      {$I-} Readln( IniFile, IniLine ); {$I+}  IniCode := IOResult;
      If IniCode = 0
        THEN BEGIN  { read line ok }
          IniLine := Upper( IniLine );  Strip( IniLine );
          RawIniLine := IniLine;
          If IniLine <> ''
            THEN BEGIN  { valid line }
              Parse := Pos( ' ', IniLine );
              State := Copy( IniLine, 1, Parse - 1 );
              Arg := Copy( IniLine, Parse + 1, Length( IniLine ) );
              Strip( Arg );  IniWord := 0;
              For IniLoop := 1 to MaxIniWords do
                If Copy( State, 1, Length( IniWords[ IniLoop ] ) ) =
                   IniWords[ IniLoop ]
                  THEN IniWord := IniLoop;
             { Check if this IniLine is valid for this program or }
             { it contains an 'END' statement. }
              If ValidStuff or ( IniWord = 23 ) then
                Case IniWord of
                   0 : If not CheckInitLine( RawIniLine )
                         THEN InitLog( 'Unknown: "' + RawIniLine + '"' );
                   1 : InitLog( 'Comment: ' + IniLine ); { comment }
                   2 : BEGIN
                         Sys.PlantName := Arg;
                         InitLog( 'Plant: ' + Sys.PlantName );
                       END;
                   3 : IniPath( 'Soil', Sys.SoilPath );
                   4 : IniPath( 'Misc', Sys.MiscPath );
                   5 : IniPath( 'Clean', Sys.CleanPath );
                   6 : BEGIN
                         Sys.TankEnable := Arg = 'ON';
                         If Sys.TankEnable
                           THEN InitLog( 'Tank enabled' );
                       END;
                   7 : IniPath( 'Log', Sys.LogPath );
                   8 : BEGIN
                         Sys.InstallCode := PadString( Arg, '~', -3 );
                         InitLog( 'Install code: ' + Sys.InstallCode );
                       END;
                   9 : IniPath( 'Help', Sys.HelpPath );
                  10 : BEGIN
                         Sys.UseSystemDate := Arg = 'ON';
                         If Sys.UseSystemDate
                           THEN InitLog( 'System date used.' );
                       END;
                  11 : BEGIN
                         Sys.SoundOn := Arg = 'ON';
                         If not Sys.SoundOn
                           THEN InitLog( 'Sounds disabled.' );
                       END;
                  12 : IniPath( 'Comm.', Sys.CommPath );
                  13 : BEGIN
                         Sys.MinYear := ParseInt( Arg );
                         If not ConversionError Then
                           Sys.MaxYear := ParseInt( Arg );
                         If ConversionError or
                            ( Sys.MaxYear < Sys.MinYear ) or
                            ( Sys.MinYear < 1980 ) or ( Sys.MaxYear > 2079 )
                           THEN IniCode := -99;
                         ChkErr( IniCode, 'x', '', 'Bad allowed year range.' );
                         InitLog( 'Year range: ' +
                            I_S( Sys.MinYear ) + '..' + I_S( Sys.MaxYear ) );
                       END;
                  14 : BEGIN
                         Sys.SystemType := Arg;
                         InitLog( 'System type: ' + Sys.SystemType );
                       END;
                  15 : BEGIN
                         Sys.ComputerName := Arg;
                         InitLog( 'Terminal name: ' + Sys.ComputerName );
                       END;
                  16 : BEGIN
                         Sys.EuropeanDate := ParseInt( Arg ) = 1;
                         If Sys.EuropeanDate
                           THEN InitLog( 'European date format.' );
                       END;
                  17 : BEGIN
                         Sys.Silent := Arg = 'ON';
                         If Sys.Silent
                           THEN InitLog( 'Silent, logs inhibited.' );
                       END;
                  18 : BEGIN
                         Sys.KeyLogEnabled := Arg = 'ON';
                         If Sys.KeyLogEnabled
                           THEN InitLog( 'Key log enabled.' );
                       END;
                  19 : BEGIN
                         Sys.ScreenSaverTime := ParseInt( Arg );
                         InitLog( 'Screen saver time ' +
                                 I_S( Sys.ScreenSaverTime ) );
                       END;
                  20 : If not Sys.IgnoreLowMemory
                         THEN IniCode := DiskFreeCheck( Arg );
                  21 : BEGIN
                         OutDev := Upper( Arg );
                         InitLog( 'Output device = ' + OutDev );
                       END;
                  22 : BEGIN    {'BEGIN'}
                         If Upper(Arg) <> Upper(Sys.ProgramName) then
                           ValidStuff := False;
                       END;
                  23 : BEGIN    {'END'}
                         ValidStuff := True;
                       END;
                  24 : IniPath( 'Master', Sys.DTAPath );
                  25 : IniPath( 'Data', Sys.DATPath );
                  26 : BEGIN
                         CommaStr := '.';  DecimalStr := ',';
                         InitLog( 'European number format.' );
                       END;
                  End;
            END;  { valid line }
        END;  { read line ok }
    END;  { file open and not eof }
  {$I-} Close( IniFile ); {$I+} JunkResult := IOResult;
  If FileExists( 'SOUNDOFF.FLG' ) THEN Sys.SoundOn := False;
  ChkErr( IniCode, 'r',
          IniFileSpec, 'initialization' );
  If Sys.DebugEnabled
    THEN BEGIN
      If Sys.Remote
        THEN InitLog( 'Remote enabled.' );
      Delay( 1000 );
      If KeyPressed
        THEN BEGIN  EatKeys;  Repeat Until KeyPressed;  EatKeys;  END;
    END;
  If ParamScan( '/D2' ) > 0 THEN Sys.InstallCode := '000';
END;  { ReadInit }


{--------------------------------------------------------------------------}
{ PROCEDURE TidyExit                                                       }
{                                                                          }
{ This procedure gets called whenever the system exits, through a fatal    }
{ error or a normal exit.                                                  }
{                                                                          }
{ Error 0 is standard termination.  "END"                                  }
{ Error code -99 is used for internal process errors.  "HALT"              }
{ Error 1 is a Turbo error previously trapped by ChkErr.  "HALT"           }
{ Errors 2..203 are Turbo errors.  "HALT"                                  }
{ Error 254 is screen saver.  "STOP"                                       }
{ Error 255 is Ctrl-Break.  "BREAK"                                        }
{ Program exit code is 0 for ok, 1 for any error,                          }
{--------------------------------------------------------------------------}

PROCEDURE TidyExit;

VAR  TidyAddr : Pointer;
     TidyMsg, MinRamStr, MaxStackUsedStr, ScreenStackUsedStr : GenlString;

  Procedure Vamoos( Msg : SmallString;  HaltCode : Integer );
    Begin
      If MaxStackUsedStr <> '' Then ProgramLog( MaxStackUsedStr );
      If MinRamStr <> '' Then ProgramLog( MinRamStr );
      If ScreenStackUsedStr <> '' Then ProgramLog( ScreenStackUsedStr );
      If HaltCode <> 0 Then ProgramLog( 'Last operation: ' + FullOperation );
      ProgramLog( Pad( Msg + ':', -10 ) + FullName );
      If ScreenCleared and ( not Sys.DebugEnabled ) Then FullWindow;
      Halt( HaltCode );
    End;

BEGIN
  TidyAddr := ErrorAddr;
  If TidyCode = 0 Then TidyCode := ExitCode  Else Exit;
  ExitProc := ExitSave;
  Cursor( On );
  {$I-} Close( KeyLogFile ); {$I+} JunkResult := IOResult;
  MaxStackUsedStr := '';  ScreenStackUsedStr := '';  MinRamStr := '';
  If Sys.InstallCode = '000' Then
    Begin
      MaxStackUsedStr :=
        'Max stack=' + I_S( MaxStackUsed ) + ' in ' + LowStackMsg + '.';
      MinRamStr :=
        'Min RAM=' + I_S( LowRam ) + ' in ' + LowRamMsg + '.';
      If ScreenStackPointer > 0
        THEN ScreenStackUsedStr :=
          'Unresolved screen stack = ' + I_S( ScreenStackPointer ) + '.';
    End;
  If TidyCode = 0    { normal termination }
    THEN BEGIN  Vamoos( 'END', 0 );  END;
  If TidyCode = 1    { termination with flag to batch file }
    THEN BEGIN  Vamoos( 'HALT', 1 );  END;
  If TidyCode = 253  { ChkErr trapped }
    THEN BEGIN  Vamoos( 'FATAL', 253 );  END;
  If TidyCode = 254  { screen saver timeout }
    THEN BEGIN  Vamoos( 'TIMEOUT', 254 );  END;
  If TidyCode = 255  { Ctrl-Break }
    THEN BEGIN
      AuditLog( 'User break at address ' + PtrHex( TidyAddr ) );
      Vamoos( 'BREAK', 255 );
    END;
  { crash, error codes 2..252 }
  TidyMsg := ErrorName( TidyCode ) + ' at address ' + PtrHex( TidyAddr );
  ErrorLog( TidyMsg );
  Message( TidyMsg, Msg_SirenCallHalt );
  Vamoos( 'CRASH', TidyCode );
END;


{--------------------------------------------------------------------------}
{PROCEDURE Initialize                                                      }
{                                                                          }
{ This procedure should be called at the beginning of every program.  It   }
{ starts the audit log and other logs and calls ReadInit to read the       }
{ command line and the configuration file.                                 }
{--------------------------------------------------------------------------}

PROCEDURE Initialize( NameIn, VerIn, RevIn : String40 );
  BEGIN
    Sys.ProgramName := NameIn;  Sys.Version := VerIn;  Sys.Revision := RevIn;
    ExitSave := ExitProc;  ExitProc := @TidyExit;
    RevIn := '';
    ProgramLog( 'BEGIN: ' + FullName + RevIn );
    ReadInit;
    AuditLogOpen := True;
    AuditLog( 'BEGIN: ' + FullName + RevIn );
    HelpChapter := Sys.ProgramName;  HelpDisabled := False;
    HelpFileSpec := Sys.HelpPath + Sys.ProgramName + '.HLP';
    OpenKeyLog;
  END;

{--------------------------------------------------------------------------}

{ Returns true if passed config file line is used as an extended }
{ config statement by the application program. }

Function NilCheckInitLine( InitLine : GenlString ) : Boolean;
  Begin  NilCheckInitLine := False;  End;

{--------------------------------------------------------------------------}

BEGIN

  CheckInitLine := NilCheckInitLine;
  ProgramHelp := DefaultProgramHelp;
  InitLog( 'Heap = ' + I_S( HeapStart ) + '..' + I_S( HeapEnd ) +
           ' = ' + I_S( HeapEnd - HeapStart ) );

END.

