(*********************************************************************
**********************************************************************
****                                                              ****
****                     INPUTMAN.PAS                             ****
****                                                              ****
****       Universal data entry handling.  Also password.         ****
****                                                              ****
**********************************************************************
**********************************************************************)


UNIT InputMan;


{$I LIBRARY.CFG }


INTERFACE


Uses Dos, Crt,
     Global, Sounds, Math, Strings, Video, Time, Key, Disk, Init;


{------------------  UNIVERSAL INPUT GLOBALS  ----------------------}

CONST
  EnterPrompt = ' > ';
  InputUnknownChar = #168;  { used in display }
  InputPadChar = #250;  { used in display }

{ input manager configuration and flags }
  RangeMin     : Real = 0.0;
  RangeMax     : Real = 0.0;
  MaxDecimal   : Byte = 4;         { max digits ever shown right of decimal }
  Changed      : Boolean = False;  { true if last call changed field }
  AnyChanged   : Boolean = False;  { true after any change made ever }
  ForceUC      : Boolean = False;  { force upper case }
  InsertMode   : Boolean = True;   { insert / overwrite mode }
  ShowDots     : Boolean = True;   { show dots for field width }
  AutoEnter    : Boolean = True;   { accept when field width filled }
  HideInput    : Boolean = False;  { hide typing, as in for password }
  PreKey       : Char = NoKey;     { pass in char as if first key typed }


FUNCTION  NumericVar( VarType : Integer ) : Boolean;
FUNCTION  VarWidth( VarType : Integer ) : Integer;
FUNCTION  VarName( VarType : Integer ) : GenlString;
FUNCTION  VarSize( VarType : Integer ) : Integer;
FUNCTION  Show( VAR Data;  VarType : Integer ) : GenlString;
FUNCTION  PadShowString( VAR Data;  VarType : Integer;
                         PadCh : Char ) : GenlString;
FUNCTION  PadShow( VAR Data;  VarType : Integer ) : GenlString;
PROCEDURE Range( Min, Max : Real );
PROCEDURE Enter( VAR Data;  VarType : Integer );
PROCEDURE EnterLn( Msg : GenlString;  VAR Data;  VarType : Integer );
FUNCTION  PasswordOK( Msg:GenlString; Desc:SmallString; Level:Char ):Boolean;


IMPLEMENTATION

{------------------  UNIVERSAL INPUT PROCEDURES  -------------------}

{-------------------------------------------------------------------}
{ FUNCTION NumericVar                                               }
{                                                                   }
{ Returns true if the VarType is a numeric variable type.           }
{-------------------------------------------------------------------}

FUNCTION NumericVar( VarType : Integer ) : Boolean;

BEGIN
  NumericVar :=
    ( VarType = _Byte ) or ( VarType = _Integer ) or
    ( VarType = _Word ) or ( VarType = _LongInt ) or
    ( VarType = _BCDReal ) or ( VarType = _Real ) or
    ( VarType = _Real2 ) or ( VarType = _Real4 ) or
    ( VarType = _Int3 ) or
    ( VarType = _SignedInt3 ) or ( VarType = _PosInt );
END;


{-------------------------------------------------------------------}
{ FUNCTION VarWidth                                                 }
{                                                                   }
{ Returns the number of characters that the display of this         }
{ particular variable type could occupy.                            }
{-------------------------------------------------------------------}

Function VarWidth( VarType : Integer ) : Integer;
  Var
    Wid : Integer;
  Begin
    Operation( 'VarWid' );
    Wid := 0;
    If NumericVar( VarType ) Then
      Begin  { numeric vars }
        If ( RangeMin = 0.0 ) and ( RangeMax = 0.0 ) Then
          Begin
            RangeMin := VarRange[ VarType ].Min;
            RangeMax := VarRange[ VarType ].Max;
          End;
        If Abs( RangeMin ) > RangeMax Then RangeMax := Abs( RangeMin );
        If RangeMax >= 10000000.0 Then  { slow width calculation }
          Wid := Length( Show( RangeMax, _Real ) )
        Else
          Begin  { quick width calculation }
            RangeMax := Trunc( RangeMax );
            If RangeMax < 10 Then  { 1 }
              Wid := 1
            Else
              If RangeMax < 100 Then  { 12 }
                Wid := 2
              Else
                If RangeMax < 1000 Then  { 123 }
                  Wid := 3
                Else
                  If RangeMax < 10000 Then  { 1,345 }
                    Wid := 5  { includes comma }
                  Else
                    If RangeMax < 100000 Then  { 12,456 }
                      Wid := 6
                    Else
                      Wid := 7;  { 123,567 }
            If RangeMin < 0 Then Inc( Wid );  { for '-' sign }
            If VarType = _Real2 Then
              Wid := Wid + 3  { .23 }
            Else
              If VarType = _Real4 Then
                Wid := Wid + 5  { .2345 }
              Else
                If VarType = _Real Then
                  Wid := Wid + 1 + MaxDecimal;
          End;  { quick width calculation }
        VarWidth := Wid;
      End  { numeric vars }
    Else
      If ( VarType >= _String )  and  ( VarType < _Array ) Then  { strings }
        VarWidth := VarType - _String
      Else
        If VarType >= _Array Then  { array var }
          VarWidth := 12
        Else
          VarWidth := VarWidths[ VarType ];
    Operation( '' );
  End;


{-------------------------------------------------------------------}
{ FUNCTION VarName                                                  }
{                                                                   }
{ Returns name of variable type.                                    }
{-------------------------------------------------------------------}

FUNCTION VarName( VarType : Integer ) : GenlString;

VAR IdxStr : String[ 15 ];

BEGIN
  If VarType <= SimpeTypes
    THEN VarName := VarNames[ VarType ]
    ELSE If ( VarType >= _String ) and ( VarType < _Array )
           THEN BEGIN
             Str( VarType - _String, IdxStr );
             VarName := 'string[' + IdxStr + ']';
           END
           ELSE If VarType >= _Array
                  THEN VarName := 'bad type'
                  ELSE BEGIN
                    Str( VarType - _Array, IdxStr );
                    VarName := 'array[' + IdxStr + ']';
                  END;
END;


{-------------------------------------------------------------------}
{ FUNCTION VarSize                                                  }
{                                                                   }
{ Returns the number of bytes that this particular variable type    }
{ occupies.                                                         }
{-------------------------------------------------------------------}

FUNCTION VarSize( VarType : Integer ) : Integer;

BEGIN
  If VarType <= SimpeTypes
    THEN VarSize := VarSizes[ VarType ]
    ELSE If ( VarType >= _String )  and  ( VarType < _Array )
           THEN VarSize := VarType - _String + 1
           ELSE If VarType >= _Array
                  THEN VarSize := VarType - _Array
                  ELSE  VarSize := -1;
END;

{-------------------------------------------------------------------}
{ FUNCTION ValidState                                               }
{                                                                   }
{ If the state is not valid, this function returns the string '  '. }
{-------------------------------------------------------------------}

FUNCTION ValidState( InState : GenlString ) : GenlString;

BEGIN
  InState[ 1 ] := UpCase( InState[ 1 ] );
  InState[ 2 ] := UpCase( InState[ 2 ] );
  If not ( InState[ 1 ] in [ 'A'..'Z', 'a'..'z', ' ' ] )
    THEN InState[ 1 ] := ' ';
  If not ( InState[ 2 ] in [ 'A'..'Z', 'a'..'z', ' ' ] )
    THEN InState[ 2 ] := ' ';
  ValidState := InState;
END;

{-------------------------------------------------------------------}
{ FUNCTION Show                                                     }
{                                                                   }
{ Returns a formatted string for a value of any variable type.      }
{-------------------------------------------------------------------}

FUNCTION Show( VAR Data;  VarType : Integer ) : GenlString;

VAR
  BooleanShow    : Boolean absolute Data;
  ByteShow       : Byte absolute Data;
  IntShow        : Integer absolute Data;
  WordShow       : Word absolute Data;
  LintShow       : LongInt absolute Data;
  RealShow       : Real absolute Data;
  Int3Show       : Int3 absolute Data;
  SignedInt3Show : Int3 absolute Data;
  BinaryByteShow : Byte absolute Data;
  PosIntShow     : Integer absolute Data;
  StateShow      : Array [ 1..2 ] of Char absolute Data;
  BCDRealShow    : BCDRealType absolute Data;
  DateShow       : String8 absolute Data;
  StringShow     : String absolute Data;
  ArrayShow      : Byte absolute Data;
  ShowR : Real;
  ShowS : GenlString;
  Point : Byte;

BEGIN
  If VarSize( VarType ) < 0 THEN Exit;
  Operation( 'Show()' );
  ChkPtr( ShowS, SizeOf( ShowS ), 'input show' );
  FillChar( ShowS, SizeOf( ShowS ), $00 );
  ShowR := 0.0;
  If VarType = _Boolean
    THEN BEGIN
      If BooleanShow THEN ShowS := 'TRUE' ELSE ShowS := 'FALSE';
    END;
  If VarType = _Byte
    THEN ShowR := ByteShow + 0.0;
  If VarType = _Integer
    THEN ShowR := IntShow + 0.0;
  If VarType = _Word
    THEN ShowR := WordShow + 0.0;
  If VarType = _LongInt
    THEN ShowR := LintShow + 0.0;
  If VarType = _Real
    THEN ShowR := RealShow;
  If VarType = _Real2
    THEN ShowR := Fix( RealShow, 2 );
  If VarType = _Real4
    THEN ShowR := Fix( RealShow, 4 );
  If VarType = _Int3
    THEN ShowR := Int3Real( Int3Show );
  If VarType = _SignedInt3
    THEN ShowR := Int3Real( Int3Show );
  If VarType = _Binary
    THEN ShowS := ByteBinary( BinaryByteShow );
  If VarType = _PosInt
    THEN ShowR := PosIntShow + 0.0;
  If VarType = _State
    THEN BEGIN
      ShowS[ 1 ] := StateShow[ 1 ];  ShowS[ 2 ] := StateShow[ 2 ];
      ShowS[ 0 ] := #2;  ShowS := ValidState( ShowS );
    END;
  If VarType = _BCDReal
    THEN ShowR := BCDReal( BCDRealShow );
  If VarType = _Date  { make date pretty }
    THEN ShowS := Date8( DateShow );
  If ( VarType >= _String ) and ( VarType <= MaxStringType )
    THEN ShowS := StringShow;
  If ( VarType >= _Array ) and ( VarType <= MaxType )
    THEN BEGIN
      Str( VarType - _Array, ShowS );
      ShowS := 'Array[' + ShowS + ']';
    END;
  If ( ShowS = '' ) and NumericVar( VarType )
    THEN BEGIN
    { Determine significant decimal places, up to 4. }
      Str( ShowR:1:4, ShowS );
      If VarType = _Real2 Then
        ShowS := Copy( ShowS, 1, Length( ShowS ) - 2 )
      Else
        If VarType <> _Real4 Then
          Begin
            DeTrailString( ShowS, '0' );
            DeTrailString( ShowS, '.' );
          End;
      ShowS := Commas( ShowS );
      { convert decimal characters }
      For Point := 1 to Length( ShowS ) do
        If ShowS[ Point ] = '.' Then
          If DecimalStr <> '' Then
            ShowS[ Point ] := DecimalStr[ Length( DecimalStr ) ];
    END;
  Strip( ShowS );
  For Point := 1 to Length( ShowS ) do
    If not ( ShowS[ Point ] in [ ' '..'~' ] )
      THEN ShowS[ Point ] := InputUnknownChar;
  Show := ShowS;
  Operation( '' );
END;

{-------------------------------------------------------------------}
{ FUNCTION PadShowString                                            }
{                                                                   }
{ Returns a formatted string for a value of any variable type       }
{ padded with a selected fill character.                            }
{-------------------------------------------------------------------}

FUNCTION PadShowString( VAR Data;  VarType : Integer;
                        PadCh : Char ) : GenlString;
  VAR
    Wid, Wid2 : Integer;
  BEGIN
    If NumericVar( VarType )
      THEN BEGIN
        { determine maximum width given range }
        If ( RangeMin = 0.0 ) and ( RangeMax = 0.0 )
          THEN BEGIN
            RangeMin := VarRange[ VarType ].Min;
            RangeMax := VarRange[ VarType ].Max;
          END;
        Wid := Length( Show( RangeMin, _Real ) );
        Wid2 := Length( Show( RangeMax, _Real ) );
        If Wid2 > Wid Then Wid := Wid2;
      END
      ELSE Wid := 0 - VarWidth( VarType );
    PadShowString := PadString( Show( Data, VarType ), PadCh, Wid );
    Range( 0.0, 0.0 );
  End;

{-------------------------------------------------------------------}
{ FUNCTION PadShow                                                  }
{                                                                   }
{ Returns a formatted string for a value of any variable type       }
{ padded with spaces.                                               }
{-------------------------------------------------------------------}

FUNCTION PadShow( VAR Data;  VarType : Integer ) : GenlString;
  Begin
    PadShow := PadShowString( Data, VarType, ' ' );
  End;

{-------------------------------------------------------------------}
{ PROCEDURE Range                                                   }
{                                                                   }
{ This procedure sets a temporary range which replaces the default  }
{ numeric range for only the next call to Enter.                    }
{-------------------------------------------------------------------}

PROCEDURE Range( Min, Max : Real );

BEGIN
  RangeMin := Min;  RangeMax := Max;
END;

{-------------------------------------------------------------------}
{ PROCEDURE Enter                                                   }
{                                                                   }
{ This procedure allows you to enter a value of any variable type.  }
{ It calls the procedures PutCursor and ShowVal.  If the item to be }
{ entered was changed, the global variables 'Changed' and           }
{ 'AnyChanged' are set to true.  Otherwise, Changed is set false.   }
{-------------------------------------------------------------------}

PROCEDURE Enter( VAR Data;  VarType : Integer );

VAR   { input a variable of any type }
  BooleanShow    : Boolean absolute Data;
  ByteShow       : Byte absolute Data;
  IntShow        : Integer absolute Data;
  WordShow       : Word absolute Data;
  LintShow       : LongInt absolute Data;
  RealShow       : Real absolute Data;
  Int3Show       : Int3 absolute Data;
  SignedInt3Show : Int3 absolute Data;
  BinaryByteShow : Byte absolute Data;
  PosIntShow     : Integer absolute Data;
  BCDRealShow    : BCDRealType absolute Data;
  StateShow      : Array [ 1..2 ] of Char absolute Data;
  StringShow     : String absolute Data;
  ArrayShow      : Byte absolute Data;
  AbsArray       : Array [ 0..255 ] of Byte absolute Data;
  ArraySave      : Array [ 0..255 ] of Byte;
  Allowed : Set of Char;
  I       : Integer;
  R       : Real;
  Wid, Wid2 : Integer;
  Tog     : Byte;
  C       : Char;
  EditPt, AtX, AtY : Byte;
  CursorSave : Word;
  D1, D2, D3 : String2;
  Make, ValStr : GenlString;
  ValidKey, NumericType : Boolean;
  Moved, Done, InputAutoEnter, InputVarChanged, InputOK : Boolean;

CONST
  JumpSet = [ 'a'..'z', 'A'..'Z', '0'..'9', ':', '*', '?' ];
  TabSize = 6;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}
PROCEDURE PutCursor;

{ Moves the cursor to the next position at which data can be entered.}

VAR  NewPt : Integer;

BEGIN
  NewPt := EditPt - 1 + 3;
  If VarType = _Date
    THEN If EditPt > 4
           THEN NewPt := NewPt + 2
           ELSE If EditPt > 2
                  THEN Inc( NewPt );
  If Sys.Remote
    THEN Remote( Rep( ^H, Wid - NewPt + 3 ) );
  If not Sys.RemoteOnly
    THEN GotoXY( AtX + NewPt, AtY );
END;
{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}
PROCEDURE ShowVal;

{ Shows the current value of the variable being entered. }

VAR
  ShowMake, BtmSave : GenlString;
  ShowWid : Integer;  ShowLoop : Byte;

BEGIN
  Changed := False;
  ShowMake := Make;
  For ShowLoop := 1 to Length( ShowMake ) do
    If not ( ShowMake[ ShowLoop ] in Allowed )
      THEN ShowMake[ ShowLoop ] := InputUnknownChar;
  ShowWid := 0 - Wid;
  If VarType = _Date
    THEN BEGIN
      ShowMake := PadString( Date6( ShowMake ), InputPadChar, -6 );
      Insert( '-', ShowMake, 5 );
      Insert( '-', ShowMake, 3 );
    END
    ELSE BEGIN
      If ShowDots
        THEN ShowMake := PadString( ShowMake, InputPadChar, ShowWid )
        ELSE ShowMake := PadString( ShowMake, ' ', ShowWid );
    END;
  If not Sys.RemoteOnly
    THEN BEGIN
      If InsertMode
        THEN TextColor( White + Blink )
        ELSE TextColor( LightBlue + Blink );
      GotoXY( AtX, AtY );  Write( EnterPrompt );
    END;
  Remote( #13 + EnterPrompt );  LowVideo;  Disp( ShowMake );
  PutCursor;
END;
{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

BEGIN
  Operation( 'Enter()' );
  SaveBottomLn;
  BottomLn( Btm_DataEntry );
  AtX := WhereX;  AtY := WhereY;  Escape := False;
  KeyLogComment( 'input' );
  CursorSave := GetCursor;
  Cursor( On );
  REPEAT
  { save original value }
    ChkPtr( ArraySave, SizeOf( ArraySave ), 'input save' );
    Move( AbsArray, ArraySave, Sizeof( ArraySave ) );
    NumericType := NumericVar( VarType );
    If NumericType and ( RangeMin = 0.0 ) and ( RangeMax = 0.0 )
      THEN BEGIN
        RangeMin := VarRange[ VarType ].Min;
        RangeMax := VarRange[ VarType ].Max;
      END;
    If not NumericType
      THEN Wid := VarWidth( VarType )
      ELSE BEGIN
        { determine maximum width given range }
        Wid := Length( Show( RangeMin, _Real ) );
        Wid2 := Length( Show( RangeMax, _Real ) );
        If Wid2 > Wid Then Wid := Wid2;
      END;
    Make := Show( Data, VarType );
    Done := False;  Moved := False;  EditPt := 1;
    InputAutoEnter := AutoEnter and ( VarType = _Date );
{ setup allowed keypresses }
    If VarType = _Boolean
      THEN Allowed := [ ' ' ];
    If VarType = _Byte
      THEN Allowed := [ '0'..'9' ];
    If VarType = _Date
      THEN Allowed := [ '0'..'9', '-', ' ' ];
    If ( VarType = _Integer ) or ( VarType = _LongInt ) or
       ( VarType = _SignedInt3 )
      THEN Allowed := [ '0'..'9', '-', ',' ];
    If ( VarType = _Word ) or
       ( VarType = _Int3 ) or ( VarType = _PosInt )
      THEN Allowed := [ '0'..'9', ',' ];
    If ( VarType = _Real ) or ( VarType = _Real2 ) or
       ( VarType = _Real4 ) or ( VarType = _BCDReal )
      THEN Allowed := [ '0'..'9', '-', '.', 'E', 'e', ',' ];
    If VarType = _Binary
      THEN Allowed := [ '0', '1' ];
    If VarType = _State
      THEN Allowed := [ 'A'..'Z', 'a'..'z', ' ' ];
    If ( VarType >= _String ) and ( VarType <= MaxStringType )
      THEN Allowed := [ ' '..'~' ];
    If ( VarType = NoType ) or
       ( VarType >= _Array ) and ( VarType <= MaxType )
      THEN Allowed := [  ];
{ initialize }
    ShowVal;
    If VarType = _Boolean
      THEN BEGIN  Tog := ByteShow;  ShowDots := False;  END;
{ take keypresses }
    REPEAT   { assemble whole line }
      InputOK := True;
      ValidKey := False;
      If PreKey <> NoKey
        THEN C := PreKey ELSE C := GetKey;
      If ForceUC THEN C := UpCase( C );
{ valid key }
      If C in allowed
        THEN BEGIN   { allowed }
{ input toggled types }
          ValidKey := True;
          If VarType = _Boolean
            THEN BEGIN   { toggle }
              Inc( Tog );  ShowVal;  DeeDeeSound;
            END     { toggle }
            ELSE BEGIN     { assemble }
{ input assembled types }
              If ( EditPt = 1 ) and ( not Moved ) and
                 ( VarType <> _Binary )
                THEN Make := '';
              If ( VarType = _Date ) and ( ( C = '-' ) or ( C = ' ' ) )
                THEN BEGIN
                  If Sys.EuropeanDate
                    THEN BEGIN
                      D1 := Copy( DateString, 4, 2 );  { DD }
                      D2 := Copy( DateString, 1, 2 );  { MM }
                    END
                    ELSE BEGIN
                      D1 := Copy( DateString, 1, 2 );  { MM }
                      D2 := Copy( DateString, 4, 2 );  { DD }
                    END;
                  D3 := Copy( DateString, 7, 2 );
                  Case Length( Make ) of
                    0 : Make := Make + D1;
                    1 : Insert( '0', Make, 1 );
                    2 : Make := Make + D2;
                    3 : Insert( '0', Make, 3 );
                    4 : Make := Make + D3;
                    5 : Insert( D3[ 1 ], Make, 5 );
                   End;
                  EditPt := Length( Make ) + 1;
                  DitSound;
                END  { date hyphen }
                ELSE BEGIN  { not date hyphen }
                  If InsertMode and ( VarType <> _Binary ) and
                     ( EditPt <= Length( Make ) ) and
                     ( Length( Make ) < Wid )
                    THEN Insert( C, Make, EditPt )
                    ELSE Make[ EditPt ] := C;
                  If EditPt > Length( Make )
                    THEN Make[ 0 ] := Chr( EditPt );
                  If EditPt <= Wid
                    THEN BEGIN
                      DitSound;
                      If EditPt < Wid THEN Inc( EditPt );
                    END
                    ELSE OopSound;
                  If ( VarType = _Binary ) and ( EditPt = 5 )
                    THEN EditPt := 6;  { skip inter-nibble space }
                END;  { not date hyphen }
              Moved := True;  ShowVal;
              If InputAutoEnter and
                 ( ( Length( Make ) = Wid ) or
                   ( ( VarType = _Date ) and ( Length( Make ) = 6 ) ) )
                THEN C := EnterKey;
            END;      { assemble }
        END;   { allowed }
{ clear line }
      If C = ^Y THEN BEGIN  DooSound;  Make := '';  End;
{ backspace }
      If C = BackSpaceKey
        THEN BEGIN  { backspace }
          ValidKey := True;
          If ( Length( Make ) >= 1 ) and ( EditPt > 1 ) and
             ( VarType <> _Binary )
            THEN BEGIN
              Dec( EditPt );  Moved := True;
              DooSound;  Delete( Make, EditPt, 1 );
              ShowVal;
            END ELSE OopSound;
        END;  { backspace }
{ extended keys }
      If VarType <> _Date Then  { cursor keys not supported for date }
      Case C of
     TabKey : BEGIN
                ValidKey := True;
                Moved := True;  DiSound;
                If EditPt + TabSize > Length( Make )
                  THEN EditPt := Length( Make ) + 1
                  ELSE EditPt := EditPt + TabSize;
                PutCursor;
               END;
^S, LfArKey : BEGIN
                ValidKey := True;
                If EditPt > 1
                  THEN BEGIN  Dec( EditPt );  DiSound;  END
                  ELSE OopSound;
                If ( VarType = _Binary ) and
                   ( EditPt = 5 ) THEN EditPt := 4;
                Moved := True;
                PutCursor;
               END;
^D, RtArKey : BEGIN
                ValidKey := True;
                If EditPt <= Length( Make )
                  THEN BEGIN  Inc( EditPt );  DiSound;  END
                  ELSE OopSound;
                If ( VarType = _Binary ) and
                   ( EditPt = 5 ) THEN EditPt := 6;
                Moved := True;
                PutCursor;
               END;
 ^C, EndKey : BEGIN
                ValidKey := True;
                If Length( Make ) < Wid
                  THEN EditPt := Length( Make ) + 1
                  ELSE EditPt := Length( Make );
                Moved := True;  DiSound;
                PutCursor;
               END;
^R, HomeKey : BEGIN
                ValidKey := True;
                EditPt := 1;  Moved := True;  DiSound;
                PutCursor;
               END;
 ^V, InsKey : BEGIN
                ValidKey := True;
                InsertMode := not InsertMode;  Moved := True;
                ShowVal;  BleepSound;
                PutCursor;
               END;
 ^G, DelKey : BEGIN
                ValidKey := True;
                If ( Make <> '' ) and ( VarType <> _Binary )
                  THEN BEGIN  { delete }
                    Moved := True;
                    Delete( Make, EditPt, 1 );
                    If ( EditPt > Length( Make ) ) and ( Make <> '' )
                      THEN EditPt := Length( Make );
                    ShowVal;  DooSound;
                    PutCursor;
                  END ELSE OopSound;
               END;
^A, CtrlLfArKey :
              BEGIN
                ValidKey := True;
                If ( VarType <> _Binary ) and
                   ( Make <> '' ) and ( EditPt > 1 )
                  THEN BEGIN  { jump left }
                    Moved := True;  DiSound;
                    If EditPt > Length( Make )
                      THEN EditPt := Length( Make );
                    While ( not ( Make[ EditPt - 1 ] in JumpSet ) ) and
                          ( EditPt > 1 ) do Dec( EditPt );
                    While ( Make[ EditPt - 1 ] in JumpSet ) and
                          ( EditPt > 1 ) do Dec( EditPt );
                    PutCursor;
                  END ELSE OopSound;
               END;
^F, CtrlRtArKey :
              BEGIN
                ValidKey := True;
                If ( VarType <> _Binary ) and
                   ( Make <> '' ) and ( EditPt < Length( Make ) )
                  THEN BEGIN  { jump right }
                    Moved := True;  Dec( EditPt );  DiSound;
                    While ( Make[ EditPt + 1 ] in JumpSet ) and
                          ( EditPt <= Length( Make ) ) do Inc( EditPt );
                    While ( not ( Make[ EditPt + 1 ] in JumpSet ) ) and
                          ( EditPt <= Length( Make ) ) do Inc( EditPt );
                    If EditPt <= Length( Make ) THEN Inc( EditPt );
                    PutCursor;
                  END ELSE OopSound;
               END;
 BackTabKey : BEGIN
                ValidKey := True;
                Moved := True;  DiSound;
                If EditPt < TabSize
                  THEN EditPt := 1
                  ELSE EditPt := EditPt - TabSize;
                PutCursor;
              END;
       End;
{ escape }
      If C = EscKey
        THEN BEGIN  { undo }
          ValidKey := True;
          I := VarSize( VarType );  ZoopSound;
          ChkPtr( AbsArray, I, 'input set' );
          Move( ArraySave, AbsArray, I );
          Done := True;  InputOK := True;
          Escape := True;
        END;  { undo }
{ enter }
      If C = EnterKey
        THEN BEGIN  { accept }
          ValidKey := True;
          If VarType = _State
            THEN BEGIN
              Done := True;
              Make := ValidState( Make );
              StateShow[ 1 ] := Make[ 1 ];
              StateShow[ 2 ] := Make[ 2 ];
            END;
          If ( VarType >= _String ) and ( VarType <= MaxStringType )
            THEN BEGIN
              Done := True;  { string }
              StringShow := Make;
            END;
          If VarType = _Date
            THEN BEGIN  { date }
              If ValidDate( Make )
                THEN BEGIN
                  Done := True;
                  StringShow := Show( Make, _Date );
                END
                ELSE BEGIN
                  Message( 'BAD DATE - PLEASE ENTER AGAIN', Msg_OopKey );
                  InputOK := False;  Done := True;  Escape := False;
                END;
            END;
          If VarType = _Boolean
            THEN BEGIN
              Done := True;
              BooleanShow := Boolean( Tog mod 2 );
            END;
          If VarType = _Binary
            THEN BEGIN
              Done := True;
              BinaryByteShow := BinaryWord( Make );
            END;
          If NumericType
            THEN BEGIN  { numeric type }
              ValStr := '';
              For I := 1 to Length( Make ) do
                If Make[ I ] in [ '0'..'9', '-', '.', 'E', 'e' ]
                  THEN ValStr := ValStr + Make[ I ];
              Val( ValStr, R, I );
              If I > 0
                THEN BEGIN    { bad number }
                  Message( 'BAD VALUE - PLEASE ENTER AGAIN', Msg_OopKey );
                  InputOK := False;  Done := True;  Escape := False;
                END      { bad number }
                ELSE BEGIN    { ok number, check range }
                  If ( R < RangeMin ) or ( R > RangeMax )
                    THEN BEGIN  { out of range }
                      Message( 'VALUE IS OUT OF RANGE - ' +
                               'PLEASE ENTER AGAIN', Msg_OopKey );
                      InputOK := False;  Done := True;  Escape := False;
                    END    { out of range }
                    ELSE BEGIN  { ok range }
                      If VarType = _Byte
                        THEN ByteShow := Trunc( R );
                      If VarType = _Integer
                        THEN IntShow := Trunc( R );
                      If VarType = _Word
                        THEN WordShow := Trunc( R );
                      If VarType = _LongInt
                        THEN LintShow := Trunc( R );
                      If VarType = _Real
                        THEN RealShow := R;
                      If VarType = _Real2
                        THEN RealShow := Fix( R, 2 );
                      If VarType = _Real4
                        THEN RealShow := Fix( R, 4 );
                      If VarType = _BCDReal
                        THEN RealBCD( R, BCDRealShow );
                      If VarType = _Int3
                        THEN RealInt3( R, Int3Show );
                      If VarType = _SignedInt3
                        THEN RealInt3( R, SignedInt3Show );
                      If VarType = _PosInt
                        THEN PosIntShow := Trunc( R );
                      Done := True;
                    END;   { ok range }
                END;     { ok number, check range }
            END;   { numeric type }
          If InputOK THEN EnterSound;
        END;   { accept case }
      PreKey := NoKey;
      If not ValidKey THEN OopSound;
    UNTIL  Done;
  UNTIL InputOK;
  ForceUC := False;  ShowDots := True;
  InputVarChanged := False;
  For EditPt := 0 to VarSize( VarType ) - 1 do
    If AbsArray[ EditPt ] <> ArraySave[ EditPt ]
      THEN InputVarChanged := True;
  Changed := InputVarChanged;
  If Changed THEN AnyChanged := True;
  RemoteLn( '' );
  Range( 0.0, 0.0 );
  BottomLn( BottomLnSave );
  Operation( '' );
  SetCursor( CursorSave );
END;  { Enter procedure }


{--------------------------------------------------------------------}
{ Procedure EnterLn                                                  }
{                                                                    }
{ Show message window with prompt, enter a variable.                 }
{--------------------------------------------------------------------}

Procedure EnterLn( Msg : GenlString;  VAR Data;  VarType : Integer );
  Var
    Wid : Integer;
  Begin
    Wid := Length( EnterPrompt + PadShow( Data, VarType ) );
    Msg := Msg + Rep( ' ', Wid );
    Message( Msg, Msg_PlainPush );
    Disp( Rep( ^H, Wid ) );
    Enter( Data, VarType );
    PopScreen;
  End;


{----------------  END OF UNIVERSAL TYPE HANDLING  ------------------}

{--------------------------------------------------------------------}
{ FUNCTION  PasswordOK                                               }
{                                                                    }
{ Prompts the user for a password and checks if that password is in  }
{ the password file and therefore valid. Return true if the password }
{ is valid.                                                          }
{--------------------------------------------------------------------}

FUNCTION PasswordOK( Msg:GenlString; Desc:SmallString; Level:Char ):Boolean;

CONST  MaxPWLen = 6;

VAR  PWSpec, PWRead : GenlString;
     PWFile : Text;  Found : Boolean;  PWErr : Integer;
     PWIn : String[ MaxPWLen ];

BEGIN
  Found := False;  PasswordOK := False;  PWIn := '';
  Level := UpCase( Level );
  If Msg = '' THEN Msg := 'Enter password';
  Msg := Msg + EnterPrompt + Rep( InputPadChar, _String + MaxPWLen );
  Message( Msg, Msg_PlainPush );
  HideInput := True;
  Enter( PWIn, _String + MaxPWLen );
  If Escape then
    BEGIN  PopScreen;  Exit;  END;
  PWIn := Upper( PWIn );
  PWSpec := Sys.MiscPath + 'PASSWORD.DTA';
  Assign( PWFile, PWSpec);
  {$I-} Reset( PWFile ); {$I+}  PWErr := IOResult;
  While ( PWErr = 0 ) and ( not Eof( PWFile ) ) and ( not Found ) Do
    BEGIN  { read loop }
      {$I-} Readln( PWFile, PWRead ); {$I+}  PWErr := IOResult;
      If PWErr = 0
        THEN BEGIN  { read ok }
          Strip( PWRead );  PWRead := Upper( PWRead );
          If Length( PWRead ) >= 3
            THEN BEGIN  { valid password file line }
              If PWRead[ 1 ] = Level
                THEN BEGIN  { correct level }
                  Delete( PWRead, 1, 2 );
                  PWRead := Copy( PWRead, 1, MaxPWLen );
                  PWRead := Upper( PWRead );
                  If PWRead = PWIn THEN Found := True;
                END;  { correct level }
            END;  { valid password file line }
        END;  { read ok }
    END;  { read loop }
  {$I-} Close( PWFile ); {$I+}  JunkResult := IOResult;
  PopScreen;
  If Found then
    BEGIN
      Message( 'PASSWORD ACCEPTED', Msg_PlainPush );  EnterSound;
      AuditLog( 'Password "' + PWIn + '" OK, ' + Desc );
      Delay( 1000 );  PopScreen;
    END
  ELSE
    BEGIN
      If PWErr = 0 then
        BEGIN
          Message( 'PASSWORD DENIED', Msg_PlainPush );  ClaxonSound;
          AuditLog( 'Password "' + PWIn + '" invalid, ' + Desc );
          Delay( 2000 );  PopScreen;
        END
      ELSE
        ChkErr( PWErr, 'o', '','Cannot access password file, ' + Desc );
    END;
END;

{---------------------------------------------------------------------------}

BEGIN  { initialization }
END.


