(***********************************************************************
************************************************************************
****                                                                ****
****                          MATH.PAS                              ****
****                                                                ****
****          Bit access, and math for all non basic types.         ****
****                                                                ****
************************************************************************
***********************************************************************)


Unit Math;


{$I LIBRARY.CFG }


INTERFACE


Uses Dos, Crt, Global, Sounds;

type
  Int3 = Array [ 1..3 ] of Byte;      { three byte unsigned Integer }
  TwoWordType = Array [ 1..2 ] of Word;
  BCDRealType = Array [ 0..9 ] of Byte;

const
  _Equal : Integer = 0;
  _Lower : Integer = 1;
  _Greater : Integer = 2;
  ZeroInt3 : Int3 = ($00,$00,$00);
  MaxInt3  : Int3 = ($FF,$FF,$FF);
  OneInt3  : Int3 = ($01,$00,$00);
  WordMath    : Word = 0;
  LongIntMath : LongInt = 0;
  RealMath    : Real = 0.0;
  BitMask  : Array [ 0..15 ] of Word =
      ( $0001, $0002, $0004, $0008, $0010, $0020, $0040, $0080,
        $0100, $0200, $0400, $0800, $1000, $2000, $4000, $8000 );
  LowStackMsg : GenlString = 'startup';
  LowRamMsg : GenlString = 'startup';
  ConversionError : Boolean = False;

Var
  LowStackPtr, HeapStart, HeapEnd, LowRam : LongInt;
  TempR : Real;

Function Fix( RealIn : Real; Digits : Integer ) : Real;
Function Exp10 (RealIn : Real; Exponent : Integer) : Real;
Function TestDigit ( I, Mask : LongInt) : LongInt;
Function FreeVector : Byte;
Function MaxStackUsed : LongInt;
Procedure CheckStack;
Function ModBit (I, BitNum : Word; SetBit : Boolean) : Word;
Function TestBit (I, BitNum : Integer) : Boolean;
Function Int3Real (var I : Int3) : Real;
Procedure RealInt3 (R : Real; var I : Int3);
Function Int3Int (var I : Int3) : Word;
Procedure IntInt3 (I : Word; var I3 : Int3);
Function EqualInt3 (var I1, I2 : Int3) : Boolean;
Function GreaterInt3 (Big, Small : Int3) : Boolean;
Procedure AddInt3 (var I1, I2 : Int3);
Procedure SubInt3 (var I1, I2 : Int3);
Function L_S (InNum:Int3) : SmallString;
Function PtrInt (InPtr : Pointer) : LongInt; { output string for pointer }
Function WordHex (W : Word) : SmallString;
Function PtrHex (InPtr : Pointer) : SmallString;{ output string for pointer }
Function BCDReal (BCDIn : BCDRealType) : Real;
Procedure RealBCD (RealIn : Real; var BCDOut : BCDRealType);
Function ByteBinary (B : Byte) : SmallString;
Function BinaryWord (S : SmallString) : Word;
Procedure ChkPtr (Var InPtr; PtrSize : LongInt; Msg : GenlString);


{--------------------------------------------------------------------------}

IMPLEMENTATION


(*******    FUNDAMENTAL   *********************************************)

{----------------------------------------------------------------------}
{ Function Fix                                                         }
{                                                                      }
{ returns input with decimal portion truncated to n digits             }
{ n < 0 truncates whole number portion of value.                       }
{----------------------------------------------------------------------}

Function Fix( RealIn : Real; Digits : Integer ) : Real;
  Begin
    RealIn := Exp10( RealIn, Digits );
    RealIn := Int( RealIn );
    Fix := Exp10( RealIn, 0 - Digits );
  End;




{----------------------------------------------------------------------}
{ Function Exp10                                                       }
{                                                                      }
{ returns input times 10 to the exponent power   2.9E-39..1.7E38       }
{----------------------------------------------------------------------}

Function Exp10
        (      RealIn : Real;
               Exponent : Integer
        ) : Real;

  var  ExpLoop : Byte;

  Begin
    Exp10 := 1;
    ConversionError := False;
    If Exponent = 0 Then Exit;
    ExpLoop := 1;
    While ExpLoop <= Abs( Exponent ) Do
      Begin  { exp loop }
        If Exponent < 0 Then
          Begin  { exponent < 0 }
            If Abs( RealIn ) < 2.9E-38 Then  { next loop would underflow }
              Begin  { underflow }
                ConversionError := True;
                If RealIn < 0 Then
                    Exp10 := -2.9E-39
                  Else
                    Exp10 := 2.9E-39;
                Exit;
              End;  { underflow }
            RealIn := RealIn / 10.0
          End  { exponent < 0 }
        Else
          Begin  { exponent > 0 }
            If Abs( RealIn ) > 1.7E37 Then  { next loop would overflow }
              Begin  { overflow }
                ConversionError := True;
                If RealIn < 0 Then
                    Exp10 := -1.7E38
                  Else
                    Exp10 := 1.7E38;
                Exit;
              End;  { overflow }
            RealIn := RealIn * 10.0
          End;  { exponent > 0 }
        Inc( ExpLoop )
      End;  { exp loop }
    Exp10 := RealIn;
  End;



{----------------------------------------------------------------------}
{ Function TestDigit                                                   }
{                                                                      }
{ returns value of masked digit, ie. TestDigit( 1340, 100 ) = 300      }
{ mask must be an even power of 10                                     }
{----------------------------------------------------------------------}

Function TestDigit
        (
                I,
                Mask : LongInt
        ) : LongInt;

  Begin
    TestDigit :=
       I - ( I mod Mask )  { subtract lower remainder }
        { subtract higher digits }
         - ( ( I div ( Mask * 10 ) ) * ( Mask * 10 ) );
  End;



{----------------------------------------------------------------------}
{ Function FreeVector                                                  }
{                                                                      }
{ Return first free interrupt vector starting from $72.                }
{ Halt program if none found.                                          }
{----------------------------------------------------------------------}

Function FreeVector : Byte;

  var VecLoop, OutVec : Byte;  TempAddr : Pointer;

  Begin
    VecLoop := $72;
    Repeat
      Inc( VecLoop );
      GetIntVec( VecLoop, TempAddr );
    Until ( TempAddr = NIL ) or ( VecLoop = $FF );
    If TempAddr = NIL Then
        OutVec := VecLoop
      Else
        OutVec := 0;
    FreeVector := OutVec;
    If OutVec = 0 Then
      Begin
        DispLn( NL + 'No vectors available.' + NL );  Halt;
      End;
  End;


{--------------------------------------------------------------------------}
{ FUNCTION MaxStackUsed                                                    }
{                                                                          }
{ Uses LowStackPointer to returns largest stack used in bytes.             }
{--------------------------------------------------------------------------}

Function MaxStackUsed : LongInt;

  Begin
    MaxStackUsed :=
      PtrInt( HeapOrg ) - ( LongInt( SSeg ) * $10 + LowStackPtr );
  End;


{----------------------------------------------------------------------}
{ Procedure CheckStack                                                 }
{                                                                      }
{ Simply maintains low stack pointer.                                  }
{ Use at top of procedures having greatest stack use.                  }
{----------------------------------------------------------------------}

Procedure CheckStack;

  Begin
    If LowStackPtr > SPtr Then
      Begin
        LowStackPtr := SPtr;
        LowStackMsg := FullOperation;
      End;
    If LowRam > MaxAvail Then
      Begin
        LowRam := MaxAvail;
        LowRamMsg := FullOperation;
      End;
  End;



(*******    BIT TEST AND SET   ****************************************)

{----------------------------------------------------------------------}
{ Funttion ModBit                                                      }
{                                                                      }
{ modify I, change BitNum to Set                                       }
{----------------------------------------------------------------------}

Function ModBit
        (
                I,
                BitNum : Word;
                SetBit : Boolean
        ) : Word;

  Begin
    BitNum := BitNum mod 16;
    If SetBit Then
        modBit := I or BitMask[BitNum]
    Else
        modBit := I and not BitMask[BitNum]
  End;



{----------------------------------------------------------------------}
{ Function TestBit                                                     }
{                                                                      }
{ return the value of bit BitNum in I                                  }
{----------------------------------------------------------------------}

Function TestBit
        (
                I,
                BitNum : Integer
        ) : Boolean;

  Begin
    BitNum := BitNum mod 16;
    TestBit := (I and BitMask[BitNum] ) <> 0
  End;



(*******    THREE BYTE INTEGER SUPPORT    *****************************)

{----------------------------------------------------------------------}
{ Procedure Int3Real                                                   }
{                                                                      }
{ convert a three-byte Integer into a Real                             }
{----------------------------------------------------------------------}

Function Int3Real
        (
                var I : Int3
        ) : Real;

  Begin
    Int3Real := I[1] + 256.0 * I[2] + 65536.0 * I[3]
  End;



{----------------------------------------------------------------------}
{ Procedure RealInt3                                                   }
{                                                                      }
{ convert a Real into a three-byte Integer                             }
{----------------------------------------------------------------------}

Procedure RealInt3
        (
                R : Real;
                var I : Int3
        );

  var
    L : LongInt;

  Begin
    R := Round( Abs( R ) );
    If R > Int3Real( MaxInt3 ) Then
        R := Int3Real( MaxInt3 );
    L := Round( R );
    I[3] := L div 65536;
    L := L - ( LongIntMath + I[3] ) * 65536;
    I[2] := L div 256;
    I[1] := L - ( LongIntMath + I[2] ) * 256
  End;



{----------------------------------------------------------------------}
{ Function Int3Int                                                     }
{                                                                      }
{ Convert three-byte Integer to a two byte Integer via truncation.     }
{----------------------------------------------------------------------}

Function Int3Int
        (
                var I : Int3
        ) : Word;

  Begin
    Int3Int := WordMath + I[2] * 256 + I[1]
  End;



{----------------------------------------------------------------------}
{ Procedure IntInt3                                                    }
{                                                                      }
{ Convert two byte Integer to three byte Integer.  MSByte = 0.         }
{----------------------------------------------------------------------}

Procedure IntInt3
        (
                I : Word;
                var I3 : Int3
        );

  Begin
    I3[1] := Lo( I );
    I3[2] := Hi( I );
    I3[3] := 0
  End;



{----------------------------------------------------------------------}
{ Function EqualInt3                                                   }
{                                                                      }
{ If I1 and I2 are equal Then return true Else return false.           }
{----------------------------------------------------------------------}

Function EqualInt3
        (
                var I1, I2 : Int3
        ) : Boolean;

  Begin
    EqualInt3 :=
     ( I1[1] = I2[1] ) and ( I1[2] = I2[2] ) and ( I1[3] = I2[3] );
  End;



{----------------------------------------------------------------------}
{ Function GreaterInt3                                                 }
{                                                                      }
{ True If Big > Small                                                  }
{----------------------------------------------------------------------}

Function GreaterInt3
        (
                Big,
                Small : Int3
        ) : Boolean;

  Begin
    GreaterInt3 := (Big[3] > Small[3]) or
        ((Big[3] = Small[3]) and (Big[2] > Small[2])) or
        ((Big[3] = Small[3]) and (Big[2] = Small[2]) and
            (Big[1] > Small[1]))
  End;



{----------------------------------------------------------------------}
{ Procedure AddInt3                                                    }
{                                                                      }
{ add two three-byte Integers                                          }
{----------------------------------------------------------------------}

Procedure AddInt3
        (
                var I1, I2 : Int3
        );

  var
    X : Integer;
    Sum : Integer;
    Carry : Integer;

  Begin
    Carry := 0;
    For X := 1 To 3 Do
      Begin
        Sum := I1[X] + I2[X] + Carry;
        If (Sum > 255) Then
            Carry := 1
        Else
            Carry := 0;
        I1[X] := Lo( Sum )
      End
  End;



{----------------------------------------------------------------------}
{ Procedure SubInt3                                                    }
{                                                                      }
{ subtract two three-byte Integers                                     }
{----------------------------------------------------------------------}

Procedure SubInt3
        (
                var I1, I2 : Int3
        );

  var
    R : Real;
    X : Integer;
    Sum : Integer;
    Borrow : Integer;

  Begin
    Borrow := 0;
    For X := 1 To 3 Do
      Begin
        Sum := I1[X] - I2[X] + Borrow;
        If (Sum < 0) Then
            Borrow := 1
        Else
            Borrow := 0;
        I1[X] := Lo( Sum )
      End
  End;



{----------------------------------------------------------------------}
{ Function L_S                                                         }
{                                                                      }
{ convert a three-byte Integer into a string                           }
{----------------------------------------------------------------------}

Function L_S
        (
                InNum : Int3
        ) : SmallString;

  var
    S : SmallString;

  Begin
    Str( LongInt( InNum[1] ) + LongInt( InNum[2] ) * 256 +
        LongInt( InNum[3] ) * 65536 : 0, S );
    L_S := S
  End;



(*************   POINTER SUPPORT   ************************************)

{----------------------------------------------------------------------}
{ Function PtrInt                                                      }
{                                                                      }
{ convert a pointer into a LongInt                                     }
{----------------------------------------------------------------------}

Function PtrInt
        (
                InPtr : Pointer
        ) : LongInt;  { output string for pointer }

  var
    TwoWords : TwoWordType absolute InPtr;
    IntTemp : LongInt;

  Begin
    IntTemp :=
      LongInt( TwoWords[ 2 ] ) * $10 +
      LongInt( TwoWords[ 1 ] );
    PtrInt := IntTemp;
  End;



{----------------------------------------------------------------------}
{ Function WordHex                                                     }
{                                                                      }
{ convert a word into a hex string                                     }
{----------------------------------------------------------------------}

Function WordHex
        (
                W : Word
        ) : SmallString;

  const
    Xlat : array [0..15] of Char = '0123456789ABCDEF';

  Begin
    WordHex := Xlat [ (W shr 12) and 15 ] +
        Xlat [ (W shr 8) and 15 ] +
        Xlat [ (W shr 4) and 15 ] +
        Xlat [ W and 15]
  End;



{----------------------------------------------------------------------}
{ Function PtrHex                                                      }
{                                                                      }
{ convert a pointer into a hex string                                  }
{----------------------------------------------------------------------}

Function PtrHex
        (
                InPtr : Pointer
        ) : SmallString;  { output string for pointer }

  var
    TwoWords : TwoWordType absolute InPtr;

  Begin
    If ( TwoWords[ 1 ] = 0 ) and ( TwoWords[ 2 ] = 0 ) Then
        PtrHex := '** NIL **'
    Else
        PtrHex :=
            WordHex( TwoWords[ 2 ] ) + ':' + WordHex( TwoWords[ 1 ] )
  End;



(*************   EXTENDED TYPES SUPPORT   *****************************)

{----------------------------------------------------------------------}
{ Procedure BCDReal                                                    }
{                                                                      }
{ convert a BCDRealType into a Real                                    }
{ BCD: @0 = bit 7 sign 0/1 is +/-, bits 6..0 is binary exponent + $3F) }
{      @1 .. @9 = mantissa LSB .. MSB 18 decimal nibbles               }
{----------------------------------------------------------------------}

Function BCDReal
        (
                BCDIn : BCDRealType
        ) : Real;

  var
    Exponent:Integer;
    BCDLoop:Byte;
    RealOut:Real;



    Function BCDNbl
            (
                    Nbl:Byte
            ) : Byte;

      Begin
        BCDNbl := ( ( Nbl and $F0 ) shr 4 ) * 10 + ( Nbl and $0F )
      End;



  Begin
    RealOut := 0.0;
    For BCDLoop := 5 to 9 do  { use 10 digits of the available 18 }
      Begin
        RealOut := RealOut + BCDNbl( BCDIn[ BCDLoop ] );
        If RealOut > 0.0 Then
            RealOut := RealOut / 100.0
      End;
    If BCDIn[ 0 ] and $80 <> 0 Then
        RealOut := 0.0 - RealOut;
    Exponent := Integer( BCDIn[ 0 ] and $7F ) - $3F;
    BCDReal := Exp10( RealOut, Exponent );
  End;



{----------------------------------------------------------------------}
{ Procedure RealBCD                                                    }
{                                                                      }
{ convert a Real into a BCDRealType (see BCDReal for format)           }
{----------------------------------------------------------------------}

Procedure RealBCD
        (
                RealIn : Real;
                var BCDOut : BCDRealType
        );

  var
    RealLoop, Nbl : Byte;
    Exponent : Integer;
    RlStr : SmallString;

  Begin
    ChkPtr( BCDOut, SizeOf( BCDOut ), 'math BCD fill' );
    FillChar( BCDOut, SizeOf( BCDOut ), 0 );
    Str( RealIn:16, RlStr );  { sd.mmmmmmmmmEsee }
    Delete( RlStr, 1, 1 );
    Delete( RlStr, 2, 1 );
    For RealLoop := 9 DownTo 5 do
      Begin
        Nbl := ( Ord( RlStr[1] ) - 48 ) * 10 + ( Ord( RlStr[2] ) - 48 );
        If Nbl = 0 Then
            BCDOut[ RealLoop ] := 0
        Else
            BCDOut[ RealLoop ] := ( Nbl div 10 ) * $10 + ( Nbl mod 10 );
        Delete( RlStr, 1, 2 )
      End;
    Exponent := 0;  { normalize:  1.0 > X >= 0.1 }
    While ( Abs( Exponent ) < 38 ) and
          ( ( RealIn >= 1.0 ) or ( RealIn < 0.1 ) ) do
        If RealIn >= 1.0 Then
          Begin
            RealIn := RealIn / 10.0;
            Inc( Exponent )
          End
        Else
          Begin
            RealIn := RealIn * 10.0;
            Dec( Exponent )
          End;
    BCDOut[ 0 ] := Exponent + $3F;
    If RealIn < 0.0 Then
        BCDOut[ 0 ] := BCDOut[ 0 ] or $80;
  End;



{----------------------------------------------------------------------}
{ Function ByteBinary                                                  }
{                                                                      }
{ convert a byte into a string of 1's & 0's                            }
{----------------------------------------------------------------------}

Function ByteBinary
        (
                B : Byte
        ) : SmallString;

  const
    Xlat : Array[ Boolean ] of Char = '01';

  Begin
    ByteBinary :=
        Xlat[ (B and 128) <> 0 ] +
        Xlat[ (B and 64) <> 0 ] +
        Xlat[ (B and 32) <> 0 ] +
        Xlat[ (B and 16) <> 0 ] + ' ' +
        Xlat[ (B and 8) <> 0 ] +
        Xlat[ (B and 4) <> 0 ] +
        Xlat[ (B and 2) <> 0 ] +
        Xlat[ (B and 1) <> 0 ]
  End;



{----------------------------------------------------------------------}
{ Function BinaryWord                                                  }
{                                                                      }
{ convert a string of 1's & 0's into a Word                            }
{----------------------------------------------------------------------}

Function BinaryWord
        (
                S : SmallString
        ) : Word;

  var
    Index : Integer;
    Total : Word;

  Begin
    Total := 0;
    For Index := 1 to Length(S) Do
      Begin
        If S[Index] = '0' Then
            Total := Total + Total;
        If S[Index] = '1' Then
            Total := Total + Total + 1;
      End;
    BinaryWord := Total
  End;



{----------------------------------------------------------------------}
{ Function ChkPtr                                                      }
{                                                                      }
{ Range check a pointer, always fatal if failed.                       }
{ Range is start of globals, through stack, through end of heap.       }
{----------------------------------------------------------------------}

Procedure ChkPtr
        (
                Var InPtr;
                PtrSize : LongInt;
                Msg : GenlString
        );
  Begin
    Msg := ', ' + Msg + ' in ' + FullOperation;
    If PtrSize < 1 Then
      ChkErr( -100, 'x', Msg, '' );
    If PtrSize > 64000 Then
      ChkErr( -101, 'x', Msg, '' );
    If PtrInt( Ptr( Seg( InPtr ), Ofs( InPtr ) ) ) < HeapStart Then
      ChkErr( -102, 'x', Msg, '' );
    If PtrInt( Ptr( Seg( InPtr ), Ofs( InPtr ) ) ) + PtrSize >= HeapEnd Then
      ChkErr( -103, 'x', Msg, '' );
  End;

{--------------------------------------------------------------------------}

Begin
  LowStackPtr := SPtr;
  LowRAM := MaxAvail;
  HeapStart := DSeg * $10;  { start of variables }
  HeapEnd := PtrInt( HeapOrg ) + MemAvail;
End.


