(*********************************************************************
**********************************************************************
****                                                              ****
****                        PRINT.PAS                             ****
****                                                              ****
****        Printer configuration load and status checking.       ****
****                                                              ****
**********************************************************************
**********************************************************************)

Unit Print;

{$I LIBRARY.CFG}

INTERFACE

Uses
   Dos, Crt, Global, Sounds, Math, Strings, Video, Time, Key, Disk, Init,
   InputMan, MenuMan;

CONST
  MaxHeaderLines = 3;
  MaxFooterLines = 1;
  PrvwIniSpec = 'PRVWATTR.INI';
  PrinterIniSpec = 'PRINTER.INI';
  MaxPrvwWidth = 150;
  { preview support }
  PrvwLength : Integer = 0;       { # of lines in printing file }
  PrvwWidth : LongInt = 0;        { # of chars in longest line }
  PrvwRow : LongInt = 1;
  PrvwCol : LongInt = 1;  { top left of current window in to file }
  { PrvwWaitForDocFinished is used if jump to end of doc key pressed }
  { before the preview file has been completed by the printing procedures.
  }
  PrvwWaitForDocFinished : Boolean = False;
  PrvwFileSpec = '.\SCREEN.$$$';

TYPE
  PrnCtrlBlock = RECORD
    InitLoaded    : Boolean;  { control, not loaded }
    PrinterName   : String20;
    LineSize      : Real;
    PtrInit       : String10;
    Pitch10CPI    : String10;
    Pitch12CPI    : String10;
    Pitch17CPI    : String10;
    DoubleWide    : String10;
    NoDoubleWide  : String10;
    Compress      : String10;
    Uncompress    : String10;
    Underline     : String10;
    NoUnderline   : String10;
    RibbonBlack   : String10;
    RibbonRed     : String10;
    SixLPI        : String10;
    EightLPI      : String10;
    GraphicsOn    : String10;
    GraphicsOff   : String10;
    GraphicLF     : String10;
    Bold          : String10;
    NoBold        : String10;
   End;

PgRecType =
  Record    { named for WordStar dot commands }
    pl  : Integer;    { total available page lines }
    mt  : Integer;    { blank lines above header }
    he  : Array [ 1..MaxHeaderLines ] of String150;
     { Header lines, include any imbedded NL's, but }
     { an NL at the end of a non-blank line is assumed. }
     { Blank header lines are not printed. }
    hl  : Integer;    { calculated number of header lines }
    hm  : Integer;    { blank lines after header }
    bl  : Integer;    { calculated number of body text lines }
    fm  : Integer;    { blank lines before footer }
    fo  : Array [ 1..MaxFooterLines ] of String150;
     { same as .he, but less capacity }
    fl  : Integer;    { calculated number of footer lines }
    mb  : Integer;    { blank lines after footer }
    pn  : Integer;    { current page number }
    po  : Integer;    { page offset (left margin) }
    l   : Integer;    { current line number }
  End;

PrvwLineType = Array [ 1..MaxPrvwWidth ] of
  Record  Ch : Char;  Attr : Byte;  End;


VAR
  Pages : Integer;         { total number of pages}
  P : PrnCtrlBlock;        { main printer control block for use by }
                           {   application program }
  PrinterOnLine : Boolean; { flag that printer was still OK after }
                           {   most recent line was printed }
  TextOut : Text;          { output device handle for file or printer }
  PrvwOut : File of PrvwLineType; { handle for preview generation file }
  PrevOutDev : GenlString; { detects change of printer device}
  HeaderPrinted : Boolean; { detects that a new header can be defined }
  FooterPrinted : Boolean; { detects that a new footer can be defined }
  Pg : PgRecType;          { page layout control record }
  PrintLnBuffer : Array [1..255] of Char; {for print routine device driver}


Procedure Preview( PrvwPageLength : Integer );
Function  PrinterReady : Boolean;
Function  PrinterOut : Boolean;
Function  ScreenOut : Boolean;
PROCEDURE InitPrinter( Var VarP : PrnCtrlBlock;
                     PrnInitName : GenlString );
Function  PrnName : SmallString;
Procedure ChooseOutDev;
Procedure PrintHeader;
Procedure PrintFooter;
Procedure FormFeed;
Procedure HeaderLine( InLn : BigString );
Procedure FooterLine( InLn : BigString );
Function  DefaultFooter : GenlString;
Procedure StartPrinting;
Procedure StopPrinting;
Procedure LineFeed( HowMany : Integer );
Procedure PrintLn( PrnLn : BigString );


IMPLEMENTATION


{-----------------------------------------------------------------}
{ Procedure PreviewInit                                           }
{                                                                 }
{ used by StartPrinting to initialize preview system              }
{-----------------------------------------------------------------}

Procedure PreviewInit;

Begin
  PrvwLength := 0;  PrvwWidth := 0;  PrvwRow := 1;  PrvwCol := 1;
  PrvwWaitForDocFinished := False;
  TextColor( White );  TextBackground( Black );
  Window( 1, 1, ScreenWidth, ( ScreenHeight - 1 ) );  ClrScr;
End;



{-----------------------------------------------------------------}
{ Procedure Preview                                               }
{                                                                 }
{ Given the page length, this large procedure implements a screen }
{ preview of page formatted reports.  The procedure provides      }
{ horizontal and vertical panning of a viewport across a document }
{ of unlimited pages, up to 150 (MaxPrvwWidth) columns.           }
{ This procedure is called, if screen is the output device,       }
{ at every formfeed, to display each newly generated page.        }
{-----------------------------------------------------------------}

Procedure Preview( PrvwPageLength : Integer );

Const
  PrvwTabSize = 10;
  PrvwPageNum : Integer = 1;
  OkKeys : Set of Char = [ F1Key, EscKey,
                           UpArKey, ^E, DnArKey, ^X,
                           LfArKey, ^S, RtArKey, ^D,
                           CtrlUpArKey, ^W, CtrlDnArKey, ^Z,
                           CtrlLfArKey, ^F, BackTabKey,
                           CtrlRtArKey, ^A, TabKey,
                           HomeKey {^QS}, EndKey {^QD},
                           PgUpKey, ^R, PgDnKey, ^C,
                           CtrlPgUpKey {^QR}, CtrlPgDnKey {^QC} ];
  UpOK : Boolean = False;    { ok to move up ? }
  DnOK : Boolean = True;     { ok to move down ? }
  LfOK : Boolean = False;    { ok to move left ? }
  RtOK : Boolean = True;     { ok to move right ? }

Var
  PrvwFile : File of PrvwLineType;  { file containing the preview }
  LocalRedraw, RemoteRedraw, ReDraw : Boolean;     { screen needs refreshing }
  PrevRow, PrevCol : Integer;  { detects movement }
  PrvwKey, PrevPrvwKey : Char;
  PrvwResult, CloseResult : Integer;   { traps disk read/close errors }
  DocFinished : Boolean;     { does file contain complete document }
  PrvwLastRow, PrvwLastCol : Integer;  { maximum row/column start }

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}
Procedure OpenPrvwFile;

{ Scan the preview file from the current line forward to find the number }
{ of lines in the file and the number of characters in the longest line. }

Var
  LineNum, ColNum, EndCol, StartCol, CharCount : Integer;
  PrvwLine : PrvwLineType;

Begin
  DocFinished := False;
  Assign( PrvwFile, PrvwFileSpec );
  {$I-} Reset( PrvwFile ); {$I+} PrvwResult := IOResult;
  If PrvwResult <> OK Then
    Begin
      {$I-} Close( PrvwFile ); {$I+} CloseResult := IOResult;
      Exit;
    End;
  PrvwLength := FileSize( PrvwFile );
  If PrvwRow > PrvwLength Then
    Begin
      { since the preview procedure exits whenever a movement off }
      { the end of file is requested, the fact that the current line }
      { is past the end of file clearly indicates that the calling }
      { print routine has and will be preparing no more pages. }
      DocFinished := True;
      If PrvwWaitForDocFinished Then
        PrvwWaitForDocFinished := False;
      If PrvwLength <= ScreenHeight - 2 Then
        PrvwRow := 1
      Else
        PrvwRow := PrvwLength - ( ScreenHeight - 1 );
  End
Else
  Begin
    { search rest of file for longest line }
    {$I-} Seek( PrvwFile, PrvwRow - 1 ); {$I+}
    PrvwResult := IOResult;
    LineNum := PrvwRow;
    EndCol := PrvwWidth;
    While ( PrvwResult = OK ) and ( LineNum <= PrvwLength ) do
      Begin  { open ok }
        {$I-} Read( PrvwFile, PrvwLine ); {$I+}
        PrvwResult := IOResult;
        If PrvwResult = OK then
          Begin  { read ok }
            StartCol := PrvwWidth;
            If StartCol < 1 Then StartCol := 1;
            For ColNum := StartCol to MaxPrvwWidth do
              If ( PrvwLine[ ColNum ].Ch in PrintableCharSet ) and
                 ( PrvwLine[ ColNum ].Ch <> ' ' ) and
                 ( ColNum > PrvwWidth ) Then
                EndCol := ColNum;  { find last non-space }
            CharCount := 0;
            For ColNum := 1 to EndCol do
              If PrvwLine[ ColNum ].Ch in PrintableCharSet then
                CharCount := ColNum;  { count printable characters }
            If CharCount > PrvwWidth then  { new line is longest? }
              PrvwWidth := CharCount;
            Inc( LineNum );
          End;  { read ok }
      End;  { open ok }
    If PrvwWaitForDocFinished Then
      begin
        PrvwRow := PrvwLength + 1;
        {$I-} Close( PrvwFile ); {$I+} CloseResult := IOResult;
      end;
  End;
End;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

Procedure DrawWindow( Row, Col : LongInt );

{Displays the current window onto the textfile on the screen.  Also }
{calculates and displays which directions of movement are currently }
{allowed.                                                           }

Var
  LineNum : Integer;
  PrvwLine : PrvwLineType;
  Arrows : SmallString;
  RemPrvwLine : BigString;

{- - - - -}
Procedure MakeRemPrvwLn;

{Take PrvwLine, chop off ending spaces, and build a string using       }
{ every second }
{element in the PvLn. (ie. use only the PvLn[i].ch fields)                }

var
  LastMRPChar, MRPLoop : integer;
  TmpStr : BigString;

begin
  MRPLoop := 1; LastMRPChar := 0; RemPrvwLine := '';
  While (MRPLoop <= MaxPrvwWidth) do
    begin
      If PrvwLine[MRPLoop].ch <> ' ' then
        LastMRPChar := MRPLoop;
      inc(MRPLoop);
    end;
  If LastMRPChar > 0 then
    For MRPLoop := 1 to LastMRPChar do
      RemPrvwLine[MRPLoop] := PrvwLine[MRPLoop].ch;
  RemPrvwLine[0] := chr(LastMRPChar);
end;

{- - - - -}

Begin
  {show the status line }
  Arrows := '';
  If UpOk Then Arrows := Arrows + UpArChar;
  If DnOk Then Arrows := Arrows + DnArChar;
  If LfOk Then Arrows := Arrows + LfArChar;
  If RtOk Then Arrows := Arrows + RtArChar;
  Arrows := Pad( Arrows, -4 );
  LineNum := PrvwRow - ( ( PrvwPageNum - 1 ) * PrvwPageLength );
  BottomLn( ' row ' + I_S( LineNum ) + '  col ' + I_S( PrvwCol ) +
            '  page ' + I_S( PrvwPageNum ) + '/' + I_S( Pages ) +
            ',  <F1> = Help,  Available moves: ' + Arrows );
  { find the position in the file }
  If KeyPressed Then
    Exit;
  {do the local redraw first}
  If (not Sys.RemoteOnly) and LocalRedraw then
    begin
      LineNum := 1;
      {$I-} Seek( PrvwFile, Row - 1 ); {$I+} PrvwResult := IOResult;
      While ( PrvwResult = OK ) and
            ( not KeyPressed ) and ( LineNum <= ScreenHeight - 1 ) do
        Begin
          {$I-} Read( PrvwFile, PrvwLine ); {$I+} PrvwResult := IOResult;
          ScreenChkPtr( Screen^, 160, 'print preview write' );
          If PrvwResult = OK Then
            Move( PrvwLine[ Col ], Screen^[ LineNum, 1 ], 160 );
          Inc( LineNum );
        End;  { end for linenum loop }
      { done if not interrupted }
      LocalReDraw := KeyPressed or ( PrvwResult <> OK );
    end
  else
    LocalRedraw := False;
  If KeyPressed Then
    Exit;
  {If not interrupted, do the remote redraw}
  If (Sys.Remote and RemoteRedraw) then
    begin
      LineNum := 1;
      {$I-} Seek( PrvwFile, Row - 1 ); {$I+} PrvwResult := IOResult;
      While ( PrvwResult = OK ) and
            ( not KeyPressed ) and ( LineNum <= ScreenHeight - 1 ) do
        Begin
          {$I-} Read( PrvwFile, PrvwLine ); {$I+} PrvwResult := IOResult;
          If PrvwResult = OK Then
            begin
              MakeRemPrvwLn;  RemoteForceNL := True;
              Remote(RemPrvwLine);
            end;
          Inc( LineNum );
        End;  { end for linenum loop }
      { done if not interrupted }
      RemoteReDraw := KeyPressed or ( PrvwResult <> OK );
    end
  else
    RemoteRedraw := False;
{ add remote screen update here }
End;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ Show all the keys valid in Preview and their functions.  Allows the }
{ system help to function normally within this procedure. }

Procedure PrvwHelp;

Var
  ProgHelp : Boolean;

Begin
  LocalRedraw := True; RemoteRedraw := True;
  SpecialSound;
  Cursor(Off);
  Frame( 7, 4, 73, 20);
  WindowTitle(0, FrameY1, MessageTextColor , 'PREVIEW HELP');
  DispLn( '' );
  DispLn('ESC       - quit preview ' +
         '        |  F1      - program help ');
  DispLn(UpArChar + '         - up 1 line    ' +
         '        |  ' + DnArChar + '       - down 1 line ');
  DispLn(LfArChar + '         - left 1 column' +
         '        |  ' + RtArChar + '       - right 1 column');
  DispLn('Ctrl- ' + UpArChar + '   - up 10 lines' +
         ' *        |  Ctrl- ' + DnArChar + ' - down 10 lines  *');
  DispLn('Ctrl- ' + LfArChar + '   - left 10    ' +
         ' *        |  Ctrl- ' + RtArChar + ' - right 10 cols  *');
  DispLn('HOME      - left side of page' +
         '    |  END     - right side of page');
  DispLn( '' );
  DispLn('PgUp      - up 1 page');
  DispLn('PgDn      - next page');
  DispLn('Ctrl-PgUp - top of document      * ');
  DispLn('Ctrl-PgDn - bottom of document   *');
  DispLn( '' );
  DispLn('*  Press second key while holding down <Ctrl> key.');
  {ProgHelp := False;}

  WaitKeyPress;
End;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ main block of preview procedure }

Begin
{ add one line for the inserted page break mark }
  Inc( PrvwPageLength );
{ initialize variables }
  PrvwResult := OK;
  LocalReDraw := True; RemoteRedraw := True;
  PrevPrvwKey := NoKey;
{ scan the file to see how many lines and length of longest line }
  If not FileExists( PrvwFileSpec ) Then
    Exit;
  OpenPrvwFile;
  If PrvwWaitForDocFinished Then
    Exit;
  Pages := ( ( PrvwLength - 1 ) div PrvwPageLength ) + 1;
{ last possible row/column values }
  PrvwLastCol := PrvwWidth - ( ScreenWidth - 1 );
  If PrvwLength < ScreenHeight - 2 Then
    PrvwLastRow := 1
  Else
    PrvwLastRow := PrvwLength - ( ScreenHeight - 2 );
  Repeat
    { determine available directions }
    UpOK := PrvwRow > 1;
    DnOK := PrvwRow < PrvwLength - ( ScreenHeight - 2 );
    LfOK := PrvwCol > 1;
    RtOK := PrvwCol < PrvwWidth - ( ScreenWidth - 1 );
    PrvwPageNum := ( PrvwRow div PrvwPageLength ) + 1;
    Cursor( Off );  HelpDisabled := True;
    PrevPrvwKey := NoKey;
    Repeat  { get a key }
      { DrawWindow is serviced to complete any unfinished redraw }
      If LocalReDraw or RemoteRedraw Then
        DrawWindow( PrvwRow, PrvwCol );
      PrvwKey := NoKey;
      If KeyPressed Then
        PrvwKey := GetKey;
      If PrevPrvwKey = ^Q Then
        Case ControlKey( PrvwKey ) of
          ^S : PrvwKey := HomeKey;
          ^D : PrvwKey := EndKey;
          ^R : PrvwKey := CtrlPgUpKey;
          ^C : PrvwKey := CtrlPgDnKey;
         End;
      If PrvwKey = ^Q Then
        Begin
          DitSound;
          PrevPrvwKey := ^Q;
        End
      Else
        If ( PrvwResult = OK ) and
           ( PrvwKey <> NoKey ) and ( not ( PrvwKey in OKKeys ) ) Then
          Begin
            OopSound; EatKeys;
            PrevPrvwKey := NoKey;
          End;
    Until ( PrvwKey in OKKeys ) or ( PrvwResult <> OK );
    Cursor( On );
    HelpDisabled := False;
    If PrvwResult = OK Then
      Begin
        PrevRow := PrvwRow;  PrevCol := PrvwCol;
        Case PrvwKey of
          ESCKey :
            Begin  { quit }
              ZoopSound;
              QuitPrinting := True;
            End;  { quit }
          F1Key : PrvwHelp;
          UpArKey, ^E :
            If UpOK Then Dec( PrvwRow );
          DnArKey, ^X :
            If DnOK Then Inc( PrvwRow );
          LfArKey, ^S :
            If LfOK Then Dec( PrvwCol );
          RtArKey, ^D :
            If RtOK Then Inc( PrvwCol );
          CtrlUpArKey, ^W :
            If UpOk Then
              If PrvwRow > PrvwTabSize Then
                PrvwRow := PrvwRow - PrvwTabSize
              Else
                PrvwRow := 1;
          CtrlDnArKey, ^Z :
            If DnOK Then
              If PrvwRow + PrvwTabSize < PrvwLastRow Then
                PrvwRow := PrvwRow + PrvwTabSize
              Else  { far right }
                PrvwRow := PrvwLastRow;
          CtrlLfArKey, BackTabKey, ^A :
            If LfOk Then
              If PrvwCol > PrvwTabSize Then
                PrvwCol := PrvwCol - PrvwTabSize
              Else
                PrvwCol := 1;
          CtrlRtArKey, TabKey, ^F :
            If RtOK Then
              If PrvwCol + PrvwTabSize < PrvwLastCol Then
                PrvwCol := PrvwCol + PrvwTabSize
              Else  { far right }
                PrvwCol := PrvwLastCol;
          HomeKey :
            If LfOK Then PrvwCol := 1;
          EndKey :
            If RtOK Then PrvwCol := PrvwLastCol;
          PgUpKey, ^R :
            If UpOK Then
              If PrvwRow < Pg.l - 1 Then
                PrvwRow := 1
              Else
                PrvwRow := ((PrvwPageNum - 2) * Pg.l) + 1;
          PgDnKey, ^C :
            If DocFinished then
              Begin
                If DnOK Then
                  Begin
                    If PrvwPageNum = Pages Then
                      PrvwRow := PrvwLastRow
                    Else
                      PrvwRow := ( PrvwPageNum * PrvwPageLength ) + 1;
                  End;
              End
            Else
              If PrvwPageNum <> Pages then
                PrvwRow := ( PrvwPageNum * PrvwPageLength ) + 1
              Else
                PrvwRow := PrvwLength + 1;
          CtrlPgUpKey :
            PrvwRow := 1;
          CtrlPgDnKey :
            If DocFinished Then
              PrvwRow := PrvwLastRow
            Else
              Begin
                PrvwRow := PrvwLength + 1;
                PrvwWaitForDocFinished := True;
              End;
        End;  { case }
        If not ( PrvwKey in [ F1Key, EscKey ] ) Then
          If ( PrevRow = PrvwRow ) and ( PrevCol = PrvwCol ) Then
            begin
              OopSound; { no movement }
              EatKeys
            end
          Else
            Begin
              DiSound;  LocalRedraw := True; RemoteRedraw := True;
            End;
      End;  { movement keys }
  Until QuitPrinting or ( PrvwResult <> OK ) or
        ( PrvwRow > PrvwLength );  { past end of file }
  ChkErr( PrvwResult, 'R', PrvwFileSpec, 'screen preview' );
  QuitPrinting := QuitPrinting or ( PrvwResult <> 0 );
  {$I-} Close( PrvwFile ); {$I+} PrvwResult := IOResult;  { ignored }
End;  { preview procedure }



{----------------------------------------------------------------------}
{ Function PrinterReady                                                }
{                                                                      }
{ returns true if the output device is a parallel port indicating      }
{ ready status, or true if the device is not a parallel port           }
{----------------------------------------------------------------------}

FUNCTION PrinterReady:Boolean;

BEGIN
  If (OutDev = 'LPT1') or (OutDev = 'PRN') or (OutDev = 'LST') or
     (OutDev = 'LPT2') or (OutDev = 'LPT3') then
    begin
      If (OutDev = 'LPT1') or (OutDev = 'PRN') or (OutDev = 'LST') then
        Regs.DX := $0000;   {printer 0}
      If (OutDev = 'LPT2') then
        Regs.DX := $0100;   {printer 1}
      If (OutDev = 'LPT3') then
        Regs.DX := $0200;   {printer 2}
      Regs.AX := $0200;    {printer status request}
      Intr($17,Regs);
      PrinterReady :=
        ( Hi(Regs.AX) and $90 = $90 ) and ( Hi(Regs.AX) and $24 = 0 );
    end
  else PrinterReady := True;
END;



{----------------------------------------------------------------------}
{ Function PrinterOut                                                  }
{                                                                      }
{ returns true if the output device name is a possible physical        }
{ printer port, or a file with the extension ".PRN".                   }
{----------------------------------------------------------------------}

Function PrinterOut : Boolean;

Begin
  PrinterOut :=  (OutDev = 'PRN') or  (OutDev = 'LST') or
    (OutDev = 'LPT1') or (OutDev = 'LPT2') or (OutDev = 'LPT3') or
    (OutDev = 'COM1') or (OutDev = 'COM2') or (OutDev = 'COM3') or
    (OutDev = 'COM4') or
    ((Length(OutDev) > 4) and    {printer file}
      (Copy(OutDev, Length(OutDev) - 3, 4) = '.PRN'));
End;



{----------------------------------------------------------------------}
{ Function ScreenOut                                                   }
{                                                                      }
{ returns true if the output device name is screen preview             }
{----------------------------------------------------------------------}

Function ScreenOut : Boolean;

Begin
  ScreenOut := OutDev = 'CON';
End;



{----------------------------------------------------------------------}
{ Procedure InitPrinter                                                }
{                                                                      }
{ Loads a printer configuration file into a printer control block.     }
{ Displays an error message if trouble reading, but none if not found. }
{----------------------------------------------------------------------}

PROCEDURE InitPrinter( Var VarP : PrnCtrlBlock;
                     PrnInitName : GenlString );

{ Read text config file of bytes, arranged in lines as:
  NArgs, Byt1, [ Byte2, ... Byte10 ] Comment
Where each line is a control sequence for activating printer features.
The lines must be in the same fixed order as in the PrnCtrlBlock above. }

VAR
  PrnInitFile : Text;
  PrnInitResult, CloseResult : Integer;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ Reads a single line from the printer configuration file. }

FUNCTION PrnInitLine : SmallString;

VAR
  OutLn, InLn : GenlString;
  InByte, Count : Integer;

BEGIN
  PrnInitLine := '';  OutLn := '';
  If ( PrnInitResult <> OK ) or EOF( PrnInitFile ) THEN Exit;
  {$I-} ReadLn( PrnInitFile, InLn ); {$I+}
  PrnInitResult := IOResult;
  If PrnInitResult <> OK THEN Exit;
  If BlankLine( InLn ) THEN Exit;
  Count := ParseInt( InLn );
  ConversionError := ConversionError or
                     ( Count < 0 ) or ( Count > 10 );
  While ( not ConversionError ) and ( Length( OutLn ) < Count ) do
    BEGIN
      InByte := ParseInt( InLn );
      ConversionError := ConversionError or
                         ( InByte < 0 ) or ( InByte > 255 );
      If not ConversionError
        THEN OutLn := OutLn + Chr( InByte );
    END;
  If ConversionError
    THEN PrnInitResult := 106 { bad numeric format }
    ELSE PrnInitLine := OutLn;
END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{main block of Procedure InitPrinter}
Begin
  { unconditionally loads new configuration }
  PrnInitResult := OK;
  FillChar( VarP, SizeOf( VarP ), 0 );  { clear control codes }
  Assign( PrnInitFile, PrnInitName );
  {$I-} Reset( PrnInitFile ); {$I+}
  PrnInitResult := IOResult;
  IF PrnInitResult = OK
    THEN BEGIN  { open ok }
      {$I-} Readln( PrnInitFile, VarP.PrinterName ); {$I+}
      PrnInitResult := IOResult;
      InitLog( VarP.PrinterName ); { write printer type }
      If PrnInitResult = OK
        THEN BEGIN  { name ok }
          {$I-} Readln( PrnInitFile, TempSt ); {$I+}
          PrnInitResult := IOResult;
          If PrnInitResult = OK
            THEN BEGIN  { line size read ok }
              VarP.LineSize := ParseReal( TempSt );
              If ConversionError or
                 ( VarP.LineSize < 1.0 ) or ( VarP.LineSize > 17.0 )
                THEN PrnInitResult := 106;  { bad numeric format }
            END;  { line size read ok }
        END;  { name ok }
    END;  { open ok }
  VarP.PtrInit      := PrnInitLine;
  VarP.Pitch10CPI   := PrnInitLine;
  VarP.Pitch12CPI   := PrnInitLine;
  VarP.Pitch17CPI   := PrnInitLine;
  VarP.DoubleWide   := PrnInitLine;
  VarP.NoDoubleWide := PrnInitLine;
  VarP.Compress     := PrnInitLine;
  VarP.Uncompress   := PrnInitLine;
  VarP.UnderLine    := PrnInitLine;
  VarP.NoUnderLine  := PrnInitLine;
  VarP.RibbonBlack  := PrnInitLine;
  VarP.RibbonRed    := PrnInitLine;
  VarP.SixLPI       := PrnInitLine;
  VarP.EightLPI     := PrnInitLine;
  VarP.GraphicsOn   := PrnInitLine;
  VarP.GraphicsOff  := PrnInitLine;
  VarP.GraphicLF    := PrnInitLine;
  VarP.Bold         := PrnInitLine;
  VarP.NoBold       := PrnInitLine;
  {$I-} Close( PrnInitFile ); {$I+} CloseResult := IOResult;
  If PrnInitResult = 0 THEN PrnInitResult := CloseResult;
  If PrnInitResult <> _FileNotFound
    THEN ChkErr( PrnInitResult, 'O', PrnInitName, 'printer config' );
  If PrnInitResult <> OK
    THEN FillChar( VarP, SizeOf( VarP ), 0 );  { clear control codes }
  If ScreenOut and ( PrnInitResult <> OK )
    THEN BEGIN
      { default screen attributes }
      VarP.Bold := ^F + Chr( White );  { bright foreground }
      VarP.NoBold := ^F + Chr( Yellow );  { normal foreground }
      VarP.RibbonRed := ^F + Chr( Red );  { bright foreground }
      VarP.RibbonBlack := ^F + Chr( Yellow );  { normal foreground }
      VarP.DoubleWide := ^B + Chr( LightGray ); { reverse background }
      VarP.NoDoubleWide := ^B + Chr( Black ); { normal background }
      VarP.UnderLine    := ^F + Chr( Blue );  { underlined foreground }
      VarP.NoUnderLine  := ^F + Chr( Yellow );  { normal foreground }
    END;
  VarP.InitLoaded := True;
  PrevOutDev := OutDev;
End;



{----------------------------------------------------------------------}
{ Procedure PrnName                                                    }
{                                                                      }
{ returns a simplified, user friendly name for the output device       }
{----------------------------------------------------------------------}

Function PrnName : SmallString;

Begin
  If PrinterOut then
    PrnName := 'printer'
  else
    If ScreenOut then
      PrnName := 'preview file'
    else
      PrnName := 'output file';
End;



{----------------------------------------------------------------------}
{ Procedure ChooseOutDev                                               }
{                                                                      }
{ pop up a menu for output device selection                            }
{----------------------------------------------------------------------}

Procedure ChooseOutDev;

begin
  QuitPrinting := False;
  Process( 'Select output device for reports:', '' );
  QuickMenuLine( 'Screen ');
  QuickMenuLine( 'Printer');
  QuickMenuLine( 'File   ');
  BottomLn( Btm_QuickMenu );
 {get menu selection}
  Repeat
    Choice := QuickMenu( '' );
  Until Escape or
        ( ( Choice <> 0 ) and
          ( ( MenuKey in PrintableCharSet ) or ( MenuKey = EnterKey ) ) );
  If Escape then
    QuitPrinting := True
  else
    Case Choice of
      1 : OutDev := 'CON';
      2 : OutDev := 'PRN';
      3 : begin
            OutDev := Sys.ProgramName + '.LST';
            Message( 'Enter output file name:' + NL +
                     Rep( ' ', 45 ), Msg_PlainPut );
            GotoXY( WhereX - 45, WhereY );
            Enter( OutDev, _String + 40 );
            QuitPrinting := Escape;
          end;
    end;  {case}
end;



{----------------------------------------------------------------------}
{ Procedure RawPrintLn                                                 }
{                                                                      }
{ Outputs text to device, used internally by PrintLn.                  }
{ Assumes that string is one line, and adds an NL at the end.          }
{ Handles formatting for preview file.                                 }
{----------------------------------------------------------------------}

Procedure RawPrintLn( RawLn : BigString );

Var
  RawPt, PrvwPt, RawResult : Integer;
  PrvwLn : PrvwLineType;
  BackAttr, ForeAttr : Byte;
  AttrCh : Char;

Begin
  If not PrinterOnLine then Exit;
  If ScreenOut then
    Begin
      BackAttr := Black;  ForeAttr := Yellow;
      RawPt := 1;  PrvwPt := 1;
      While PrvwPt <= MaxPrvwWidth do
        Begin  { create a preview line including attributes }
          If RawPt > Length( RawLn ) then  { pad }
            Begin
              PrvwLn[ PrvwPt ].Ch := ' ';
              PrvwLn[ PrvwPt ].Attr := Black;
              Inc( PrvwPt );
            End
          else
            Begin  { characters or attributes }
              If ( RawPt < Length( RawLn ) ) and
                 ( RawLn[ RawPt ] in [ ^B, ^F ] ) then
                Begin  { attributes }
                  If RawLn[ RawPt ] = ^B  { background }
                    then BackAttr := Ord( RawLn[ RawPt + 1 ] )
                  else  { foreground }
                    ForeAttr := Ord( RawLn[ RawPt + 1 ] );
                  Inc( RawPt );
                End
              else
                Begin  { not attributes }
                  PrvwLn[ PrvwPt ].Ch := RawLn[ RawPt ];
                  PrvwLn[ PrvwPt ].Attr := BackAttr * $10 + ForeAttr;
                  Inc( PrvwPt );
                End;
              Inc( RawPt );
            End { characters or attributes }
        End;  { create preview line }
      {$I-} Write( PrvwOut, PrvwLn ); {$I+} RawResult := IOResult;
      ChkErr( RawResult, 'W', OutDev, PrnName );
    End  { preview device }
  else
    Begin  { text device }
      RawLn := RawLn + NL;
      {$I-} Write( TextOut, RawLn ); {$I+}
      RawResult := IOResult;
      ChkErr( RawResult, 'W', OutDev, PrnName );
    End;  { text device }
  Inc(Pg.l);   { increment current line number }
  PrinterOnLine := RawResult = OK;
End;



{----------------------------------------------------------------------}
{ Procedure LineFeed                                                   }
{                                                                      }
{ outputs repeated blank lines, used internally to printing system     }
{----------------------------------------------------------------------}

Procedure LineFeed( HowMany : Integer );

var
  LFLoop : integer;

begin
  For LFLoop := 1 to HowMany do
    RawPrintLn( '' );
end;



{----------------------------------------------------------------------}
{ Procedure PrintBlock                                                 }
{                                                                      }
{ Prints a header or footer, blocks known not to cause page breaks.    }
{ Used internally by PrintHeader and PrintFooter.                      }
{----------------------------------------------------------------------}

Procedure PrintBlock( TmpLn : BigString );

Var PLn : BigString;

Begin
  While TmpLn <> '' do
    begin
      ParseCharSet := [LF];
      PLn := ParseStr(TmpLn);
      If ( PLn <> '' ) and ( PLn[ Length( PLn ) ] = CR ) then
        Delete( PLn, Length( PLn ), 1 );
      RawPrintLn( PLn );
    end;
End;



{----------------------------------------------------------------------}
{ Procedure PrintHeader                                                }
{                                                                      }
{ Performs all necessary steps to print a page top.                    }
{ Used internally by PrintLn.                                          }
{----------------------------------------------------------------------}

Procedure PrintHeader;

Var HLoop : Byte;  OpenResult : Integer;

Begin
  If QuitPrinting or ( not PrinterOnLine ) then Exit;
{ open output file at top of each page, close at end }
  If ScreenOut then
    Begin
      Assign( PrvwOut, PrvwFileSpec );
      {$I-} Reset( PrvwOut ); {$I+} OpenResult := IOResult;
      If OpenResult = OK then
        Begin  { append }
          {$I-} Seek( PrvwOut, FileSize( PrvwOut ) ); {$I+}
          OpenResult := IOResult;
        End;
      If OpenResult <> OK then
        Begin
          {$I-} ReWrite( PrvwOut ); {$I+} OpenResult := IOResult;
        End;
    End
  else
    Begin
      Assign( TextOut, OutDev );
      {$I-} Append( TextOut ); {$I+} OpenResult := IOResult;
      If OpenResult <> OK then
        Begin
          {$I-} ReWrite( TextOut ); {$I+} OpenResult := IOResult;
        End;
    End;
  ChkErr( OpenResult, 'O', OutDev, PrnName );
  QuitPrinting := OpenResult <> OK;
  If QuitPrinting then Exit;
  BottomLn( 'Please wait . . .      <ESC> interrupts' );
  If ScreenOut then
    Message( 'PREPARING page ' + I_S( Pg.pn ) + ' . . .', Msg_PlainPut )
  else
    Message( 'PRINTING page ' + I_S( Pg.pn ) + ' . . .', Msg_PlainPut );
  LineFeed(Pg.mt);  { advance to start of header }
  For HLoop := 1 to MaxHeaderLines do
    PrintBlock( Pg.he[ HLoop ] );  { print header }
  LineFeed(Pg.hm);  { advance to start of text body }
  HeaderPrinted := True;
  FooterPrinted := False;
End;



{----------------------------------------------------------------------}
{ Procedure PrintFooter                                                }
{                                                                      }
{ Performs all necessary steps to print an end of page                 }
{ Used internally by PrintLn, but also useable externally              }
{ for forced form feeds.                                               }
{----------------------------------------------------------------------}

Procedure PrintFooter;

Var FLoop : Byte;

Begin
  If QuitPrinting or ( not PrinterOnLine ) then
    Exit;
  LineFeed(Pg.pl - Pg.mb - Pg.fl - Pg.l);  {skip to start of footer}
  For FLoop := 1 to MaxFooterLines do
    PrintBlock( Pg.fo[ FLoop ] );
  LineFeed( Pg.mb );
  FooterPrinted := True;
  HeaderPrinted := False;
  If ScreenOut then
    RawPrintLn( Rep( #205, ScreenWidth ) );
  Pg.l := 0;  Inc( Pg.pn );
  {$I-} Close( TextOut ); {$I+} JunkResult := IOResult;
  {$I-} Close( PrvwOut ); {$I+} JunkResult := IOResult;
  If ScreenOut then
    Preview( Pg.pl );
End;


{----------------------------------------------------------------------}
{ Procedure FormFeed                                                   }
{                                                                      }
{ Prints the footer, advanced to start of next page, and closes the    }
{ textfile.  Called internally by PrintLn but can also be called       }
{ independently.                                                       }
{----------------------------------------------------------------------}

Procedure FormFeed;

begin
 { print footer, advance to start of next page, and close textfile}
  PrintFooter;
end;


{----------------------------------------------------------------------}
{ Procedure HeaderLine                                                 }
{                                                                      }
{ Accepts one line at a time (do not imbed NL's), and assembles them   }
{ in the header storage array.                                         }
{----------------------------------------------------------------------}

Procedure HeaderLine( InLn : BigString );

Var HLoop, HSize : Byte;

begin
  If HeaderPrinted then
    Begin
      Pg.hl := 0;  FillChar( Pg.he, SizeOf( Pg.he ), 0 );
      HeaderPrinted := False;
    End;
  HLoop := 1;  HSize := SizeOf( Pg.he[ HLoop ] ) - 1;
  While ( HLoop < MaxHeaderLines ) and
        ( HSize - Length( Pg.he[ HLoop ] ) < 2 + Length( InLn ) ) do
       Inc( HLoop );
  If HSize - Length( Pg.he[ HLoop ] ) >= 2 + Length( InLn ) then
    Pg.he[ HLoop ] := Pg.he[ HLoop ] + InLn + NL;
  Inc( Pg.hl );
  Pg.bl := Pg.pl - Pg.mt - Pg.hl - Pg.hm - Pg.fm - Pg.fl - Pg.mb;
end;


{----------------------------------------------------------------------}
{ Procedure FooterLine                                                 }
{                                                                      }
{ Accepts one line at a time (do not imbed NL's), and assembles them   }
{ in the footer storage array.                                         }
{----------------------------------------------------------------------}

Procedure FooterLine( InLn : BigString );

Var FLoop, FSize : Byte;

begin
  If FooterPrinted then
    Begin
      Pg.fl := 0;  FillChar( Pg.fo, SizeOf( Pg.fo ), 0 );
      FooterPrinted := False;
    End;
  FLoop := 1;  FSize := SizeOf( Pg.fo[ FLoop ] ) - 1;
  While ( FLoop < MaxFooterLines ) and
        ( FSize - Length( Pg.fo[ FLoop ] ) < 2 + Length( InLn ) ) do
       Inc( FLoop );
  If FSize - Length( Pg.fo[ FLoop ] ) >= 2 + Length( InLn ) then
  Pg.fo[ FLoop ] := Pg.fo[ FLoop ] + InLn + NL;
  Inc( Pg.fl );
  Pg.bl := Pg.pl - Pg.mt - Pg.hl - Pg.hm - Pg.fm - Pg.fl - Pg.mb;
end;


{----------------------------------------------------------------------}
{ Function DefaultFooter                                               }
{                                                                      }
{ use with FooterLine in application program to set standard footer    }
{----------------------------------------------------------------------}

Function  DefaultFooter : GenlString;

Begin
  DefaultFooter := Sys.ProgramName +
              '      Printed: ' + DateString + ' @ ' + TimeString;
End;



{----------------------------------------------------------------------}
{ Procedure StartPrinting                                              }
{                                                                      }
{ Initializes most of the printing system, including TOF prompting.    }
{ Use after ChooseOutDev, but before setting up format, headers, ...   }
{                                                                      }
{ For the print routines to work properly, a header must be defined    }
{ after StartPrinting and before any calls to PrintLn and LineFeeds    }
{ can only come after the first PrintLn  !!!!                          }
{----------------------------------------------------------------------}


Procedure StartPrinting;

Begin
  PreviewInit;
  PrinterOnLine := False;
  FillChar( Pg, SizeOf( Pg ), 0 );  {initialize variables}
  Pg.pl  := 66;   { total available page lines }
  Pg.fm  := 2;    { blank lines above footer }
  Pg.mb  := 4;    { blank lines after footer }
  Pg.pn  := 1;    { first page }
{ if the output device is new, load the control codes }
  If ( not P.InitLoaded ) or ( OutDev <> PrevOutDev ) then
    Begin
      If PrinterOut then
        InitPrinter( P, Sys.MiscPath + PrinterIniSpec )
      else
        If ScreenOut then
          InitPrinter( P, Sys.MiscPath + PrvwIniSpec )
        else
          FillChar( P, SizeOf( P ), 0 );
     End;
 {initialize default header and footer}
  HeaderPrinted := True;  { restarts FooterLine }
  FooterPrinted := True;  { restarts FooterLine }
  PrinterOnLine := True;
  QuitPrinting := False;
  If PrinterOut then
    Begin
      Escape := False;
      While ( not PrinterReady ) and ( not Escape ) do
        Message( 'PRINTER IS NOT READY' + NL +
                 'Check printer power, and On-Line status',
                 Msg_OopKey );
      QuitPrinting := Escape;
      If not QuitPrinting then
        Begin
          Message( 'PLEASE SET PRINTER TO TOP OF FORM', Msg_AttnKey );
          QuitPrinting := WaitKeyChar = ESCKey;
        End;
    End;  { printer out }
End;



{----------------------------------------------------------------------}
{ Procedure StopPrinting                                               }
{                                                                      }
{ performs necessary printing shutdown operations                      }
{----------------------------------------------------------------------}

Procedure StopPrinting;

var
  TmpLn, PLn : BigString;

begin
 { if page has something on it, print footer and close textfile}
 If Pg.l <> 0 then
   FormFeed;
 If ScreenOut then
   While not QuitPrinting do
     Preview( Pg.pl );
 If ScreenOut then
   JunkResult := EraseFile( PrvwFileSpec );
end;



{-------------------------------------------------------------------------}
{ Procedure PrintLn                                                       }
{                                                                         }
{ Separate PrnLn into individual lines to be printed and print them.      }
{ First, check where on the page we are currently and take the            }
{ appropriate action.                                                     }
{ If Pg.l = 0 : At the top of the page, need to print the header and      }
{           advance to the line on which the body text should start.      }
{ If Pg.l = end of body text (ie., the last line): Need to advance to the }
{           start of the footer, print the footer, and advance to the     }
{           bottom of the page.  Also, the preview procedure gets called  }
{           here if necessary.                                            }
{ If Pg.l = anywhere else : we are somewhere in the body text so print    }
{           the text.                                                     }
{-------------------------------------------------------------------------}

Procedure PrintLn( PrnLn : BigString);

var
  PLn : BigString;

begin
  While PrinterOnLine and ( not QuitPrinting ) and ( PrnLn <> '' ) do
    begin
      If KeyPressed then
        Begin  { user interruption }
          If GetKey = EscKey then
            Begin
              QuitPrinting := True;  ZoopSound;
            End
          Else
            OopSound;
        End
      else  { not key pressed }
        If ( Pg.l = 0 ) and ( not HeaderPrinted ) then   {top of page}
          PrintHeader
        else
          If ( Pg.l = Pg.mt + Pg.hl + Pg.hm + Pg.bl ) and
             ( not FooterPrinted ) then {end of text body}
            PrintFooter
          else
            begin  { print body lines }
              ParseCharSet := [LF];
              PLn := ParseStr(PrnLn);
              If ( PLn <> '' ) and ( PLn[ Length( PLn ) ] = CR ) then
                Delete( PLn, Length( PLn ), 1 );
              RawPrintLn( PLn );
            End;
      If QuitPrinting then
        Begin
          {$I-} Close( TextOut ); {$I+} JunkResult := IOResult;
          {$I-} Close( PrvwOut ); {$I+} JunkResult := IOResult;
          If ScreenOut then JunkResult := EraseFile( PrvwFileSpec );
        End;
    end;  {while PrnLn <> '' loop}
end;


{----------------------------------------------------------------------}
{ Function PrnVoid                                                     }
{                                                                      }
{ Dummy function used by PrnOpen.                                      }
{----------------------------------------------------------------------}

Function PrnVoid( Var PrnFile : TextRec ) : Integer;

Begin
  PrnVoid := 0;
End;


{----------------------------------------------------------------------}
{ Function PrnClose                                                    }
{                                                                      }
{ Finishes up the print routines and closes a file that has been       }
{ assigned with a AssignPrn statement.                                 }
{----------------------------------------------------------------------}

Function PrnClose( Var PrnFile : TextRec ) : Integer;

begin
  StopPrinting;
  PrnFile.Mode := fmClosed;
  PrnClose := 0;
end;


{----------------------------------------------------------------------}
{ Function PrnWrite                                                    }
{                                                                      }
{ Gets called automatically when a file defined with an AssignPrn gets }
{ written to using a Write or Writeln.  PrnWrite breaks the contents   }
{ of the buffer into smaller lines if necessary and then prints these  }
{ lines using the library print routines (PrintLn, etc.)               }
{----------------------------------------------------------------------}

Function PrnWrite( Var PrnFile : TextRec ) : Integer;

Var
  PrnStr, PrnStr1 : BigString;

Begin
  PrnWrite := 0;
  With PrnFile do
    Begin
      If  BufPos = 0  then
        Begin  BufPos := 0;  BufEnd := 0;  Exit;  End;
      ChkPtr( PrnStr1[ 1 ], BufPos - 1, 'prn device buffer write' );
      Move( BufPtr^, PrnStr1[ 1 ], BufPos - 1 );
      PrnStr1[ 0 ] := Chr( BufPos - 1 );
      BufPos := 0;  BufEnd := 0;
      { break into shorter lines and print }
      Repeat
        ParseCharSet := [LF];
        PrnStr := ParseStr( PrnStr1 );
        If Copy( PrnStr, Length( PrnStr ), 1 ) = CR then
          Delete( PrnStr, Length( PrnStr ), 1 );
        PrintLn( PrnStr );
      Until PrnStr1 = '';
    End;
End;

{----------------------------------------------------------------------}
{ Function PrnOpen                                                     }
{                                                                      }
{ When a file (that has been previously assigned using AssignPrn) is   }
{ opened using a Rewrite or Append, this function is called            }
{ automatically.  It opens the file and initialized the library print  }
{ routines.                                                            }
{----------------------------------------------------------------------}

Function PrnOpen( Var PrnFile : TextRec ) : integer;

Begin
  With PrnFile do
    Begin
      If Mode = fmInput then
        Begin
          InOutFunc := @PrnVoid;
          FlushFunc := @PrnVoid;
        End
      else
        Begin
          Mode := fmOutput;
          InOutFunc := @PrnWrite;
          FlushFunc := @PrnWrite;
        End;
      CloseFunc := @PrnClose;
    End;
  StartPrinting;
  PrnOpen := 0;
End;


{----------------------------------------------------------------------}
{ Procedure AssignPrn                                                  }
{                                                                      }
{ Assigns an OUTPUT file to use the printer procedure PrintLn and all  }
{ its surrounding printer routines. When a file is assigned with       }
{ AssignPrn(f), whenever this file is rewritten, appended, written to, }
{ etc. the output automatically goes through the library printer       }
{ routines                                                             }
{----------------------------------------------------------------------}

Procedure AssignPrn( Var PrnFile : Text );

Begin
  With TextRec(PrnFile) do
    Begin
      Handle := $FFFF;
      Mode := fmClosed;
      BufSize := SizeOf( PrintLnBuffer );
      BufPtr := @PrintLnBuffer;
      OpenFunc := @PrnOpen;
      Name[ 0 ] := #0;
    End;
End;

{---------------------------------------------------------------------------}


BEGIN
  FillChar( P, SizeOf( P ), 0 );
END.

(*

{ Printing system, sample usage: }

Program PrnTst;

{$I LIBRARY.INC}

var
  Loop : integer;

begin

  Initialize( 'PRNTEST', '2.0', '09-27-89' );
  ClearScreen( BigTitle );
  ChooseOutDev;
  If not QuitPrinting then StartPrinting;
  HeaderLine( P.Bold + 'This is a report' + P.NoBold );
  HeaderLine( P.DoubleWide + 'about fish.' + P.NoDoubleWide );
  HeaderLine( P.Underline + 'I hope you like it.' + P.NoUnderline );
  HeaderLine( '' );
  FooterLine( DefaultFooter );
  For Loop := 1 to 70 do
    Begin
  { define continued header for subsequent pages in section, }
  { any time after first printed line, but only define once per page }
      If Pg.l = 20 then  { at 20th line }
        Begin
          HeaderLine( 'Fish report - continued.  Page ' + I_S(Pg.pn + 1) );
          HeaderLine( '' );
        End;
      PrintLn( 'This is line ' + I_S( Loop ) );
    End;
  StopPrinting;

end.


*)

