(***********************************************************************
************************************************************************
****                                                                ****
****                         TIME.PAS                               ****
****                                                                ****
****      Time and date utilities.  String and ordinal formats.     ****
****                                                                ****
************************************************************************
***********************************************************************)

(* This unit contains complete utilities for operations on times and dates
   represented as six or eight character Strings or represented
   in an absolute format LongInt, where absolute date is elapsed days
   since 1-1-1980, and absolute time is seconds since midnight.
   The "Parse" procedures convert string representations into separate
   year/month/day, hour/minute/second values.
   The "Abs" functions return LongInt values given absolute
   year/month/day, hour/minute/second.
   The "Ord" functions directly convert String to LongInt types.
   The "Str" functions directly convert LongInt to String types.
   Additional utilities are included for Next and Previous date calculations.
   An interrupt serviced timer system is implemented which allows timed
   operations such as time display updates or screen saver timeouts
   with no timer maintenance required from the foreground program.
   This is strictly hardware dependent, using interupt vector $08,
   the system timer tick service interrupt.  *)

Unit Time;


{$I LIBRARY.CFG }


INTERFACE


Uses Dos, Crt, Global, Sounds, Math, Strings, Video;

const
  DaysOfWeek : array [ 0..7 ] of String10 =
      ( 'NoDay', 'Sunday', 'Monday', 'Tuesday', 'Wednesday',
        'Thursday', 'Friday', 'Saturday' );
  Months : array [ 1..12 ] of String10 =
      ( 'January', 'February', 'March', 'April', 'May', 'June',
        'July', 'August', 'September', 'October', 'November', 'December' );
  DaysPerMonth : array [ 1..12 ] of integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  SecondsPerDay : LongInt = 86400;
  SecondServiceFreq : Word = 19;  { approx 18.2 / second }
  TickFactor : Real = 18.2065;
  FixedTimerVector = $71;
  MinTimer = -2;  MaxTimer = 5;
  ClockTimer = 0;  ScreenSaverTimer = -1;  KeyLogTimer = -2;
  PrinterReadyTimer = -1;  { shared with sceen saver }

Var
  { Timers 1..MaxTimer are utility timers for the application program. }
  { Timer 0 is used for the clock display, timer -1 is screen timeout, }
  { timer -2 is key log time stamp. }
  Timer : Array [ MinTimer..MaxTimer ] of Word;
  TimeExitSave : Pointer;
  TimerOriginalAddress : Pointer;
  TimerNewVector : Byte;
  ISRLoop : Integer;
  ISRHours, ISRMinutes, ISRSeconds : Word;


Function Date6( InDate : String8 ) : String6;
Function Date8( InDate : String8 ) : String8;
Function Str2( InNum : Integer ) : String2;
Function MakeDate( Y, M, D : Integer ) : String8;
Function MakeTime( H, M, S : Integer ) : String8;
Function TimeString : String8;
Function DateString : String8;
Function FullYear : Integer;
Function ParseDate( InDate : String8; var Y, M, D : Integer ) : Boolean;
Function ParseTime( InTime : String8; var H, M, S : Integer ) : Boolean;
Function AbsDate( Y, M, D : Integer ) : LongInt;
Function DateInDays : LongInt;
Function OrdDate( InDate : String8 ) : LongInt;
Function StrDate( InDate : LongInt ) : String8;
Function EqualDate( InDate1, InDate2 : String8 ) : Boolean;
Function GreaterDate( InDate1, InDate2 : String8 ) : Boolean;
Function AbsTime( H, M, S : Integer ) : LongInt;  { seconds since 00 : 00:00 }
Function TimeInSeconds : LongInt;
Function OrdTime( InTime : String8 ) : LongInt;
Function StrTime( InTime : LongInt ) : String8;
Function AbsDateTime( Y, Month, D, H, Minute, S : Integer ) : Real;
Function DateTimeInSeconds : Real;
Function ValidDate( InDate : String8 ) : Boolean;
Function ValidTime( InTime : String8 ) : Boolean;
Function NextDate( InDate : String8 ) : String8;
Function PrevDate( InDate : String8 ) : String8;
Function YYMMDD( InDate : String8 ) : String8;
Function DateDay( InDate : String8 ) : Byte;
Function Day( InDate : String8 ) : SmallString;
Function ShowDate( InDate : String8;  Style : Byte ) : String40;
Procedure SetTimer( TimerNum : Integer;  Duration : LongInt );
Function TimerElapsed( TimerNum : Integer ) : Boolean;


IMPLEMENTATION



{----------------------------------------------------------------------}
{ Function Date6                                                       }
{                                                                      }
{ remove all non-numeric characters from date/time strings             }
{----------------------------------------------------------------------}

Function Date6( InDate : String8 ) : String6;

  var
    RmLoop : Byte;

  Begin
    RMLoop := Length( InDate );
    While RmLoop > 0 Do
      Begin
        If not ( InDate[ RmLoop ] in [ '0'..'9' ] ) Then
            Delete( InDate, RmLoop, 1 );
        Dec( RmLoop )
      End;
    Date6 := InDate;
  End;



{----------------------------------------------------------------------}
{ Function Date8                                                       }
{                                                                      }
{ add hyphens after the second and fourth characters in a string       }
{----------------------------------------------------------------------}

Function Date8( InDate : String8 ) : String8;

  Begin
    InDate := Date6( InDate );
    Insert( '-', InDate, 3);
    Insert( '-', InDate, 6);
    Date8 := InDate;
  End;



{----------------------------------------------------------------------}
{ Function Str2                                                        }
{                                                                      }
{ convert a number 0..99 into a two character string '00'..'99'        }
{ used by TimeString & DateString                                      }
{----------------------------------------------------------------------}

Function Str2( InNum : Integer ) : String2;

  Begin
    If InNum > 99 then InNum := 99;
    Str2 := Chr( InNum div 10 + Ord( '0' ) ) +
            Chr( InNum mod 10 + Ord( '0' ) );
  End;



{----------------------------------------------------------------------}
{ Function MakeDate                                                    }
{                                                                      }
{ given individual year, month, and date, assembles a date string      }
{----------------------------------------------------------------------}

Function MakeDate( Y, M, D : Integer ) : String8;
  BEGIN
    MakeDate := Str2( M ) + '-' + Str2( D ) + '-' + Str2( Y mod 100 );
  END;



{----------------------------------------------------------------------}
{ Function MakeTime                                                    }
{                                                                      }
{ given individual hour, minute, and second, assembles a time string   }
{----------------------------------------------------------------------}

Function MakeTime( H, M, S : Integer ) : String8;
  BEGIN
    MakeTime := Str2( H ) + ':' + Str2( M ) + ':' + Str2( S );
  END;



{----------------------------------------------------------------------}
{ Function TimeString                                                  }
{                                                                      }
{ return the current system clock time in the form 'HH:MM:SS'          }
{----------------------------------------------------------------------}

Function TimeString : String8;

  var
    H, M, S, S100 : Word;

  Begin
    GetTime( H, M, S, S100 );
    TimeString := Str2( H ) + ':' + Str2( M ) + ':' + Str2( S )
  End;



{----------------------------------------------------------------------}
{ Function DateString                                                  }
{                                                                      }
{ return the current system clock date in the form 'MM-DD-YY'          }
{----------------------------------------------------------------------}

Function DateString : String8;

  var
    Y, M, D, Day : Word;

  Begin
    GetDate( Y, M, D, Day );  Y := Y mod 100;
    DateString := MakeDate( Y, M, D );
  End;



{----------------------------------------------------------------------}
{ Function FullYear                                                    }
{                                                                      }
{ return the current system year as an integer 1980..2079              }
{----------------------------------------------------------------------}

Function FullYear : Integer;

  var
    Y, M, D, Day : Word;

  Begin
    GetDate( Y, M, D, Day );  FullYear := Y;
  End;



{----------------------------------------------------------------------}
{ Function ParseDate                                                   }
{                                                                      }
{ split a date string into year, month, and day words                  }
{ return true if no numeric conversion errors, false otherwise         }
{ note that the year is returned in full four digit accuracy           }
{----------------------------------------------------------------------}

Function ParseDate( InDate : String8; var Y, M, D : Integer ) : Boolean;

  var TempOK : Boolean;

  Begin
    InDate := Date6( InDate );  TempOK := True;
    If Length( InDate ) = 6 Then
      Begin
        M := S_I( Copy( InDate, 1, 2 ) );
        If ConversionError Then TempOK := False;
        D := S_I( Copy( InDate, 3, 2 ) );
        If ConversionError Then TempOK := False;
        Y := S_I( Copy( InDate, 5, 2 ) );
        If ConversionError Then TempOK := False;
        If TempOK Then
          If Y < 80 Then Y := Y + 2000 Else Y := Y + 1900;
      End
      Else TempOK := False;
    If not TempOK Then
      Begin  Y := 0;  M := 0;  D := 0;  End;
    ParseDate := TempOK;
  End;



{----------------------------------------------------------------------}
{ Function ParseTime                                                   }
{                                                                      }
{ split a time string into hours, minutes and seconds                  }
{ return true if no numeric conversion errors, false otherwise         }
{----------------------------------------------------------------------}

Function ParseTime( InTime : String8; var H, M, S : Integer ) : Boolean;

  var TempOK : Boolean;

  Begin
    InTime := Date6( InTime );  TempOK := True;
    If Length( InTime ) = 6 Then
      Begin
        H := S_I( Copy( InTime, 1, 2 ) );
        If ConversionError Then TempOK := False;
        M := S_I( Copy( InTime, 3, 2 ) );
        If ConversionError Then TempOK := False;
        S := S_I( Copy( InTime, 5, 2 ) );
        If ConversionError Then TempOK := False;
      End
      Else TempOK := False;
    If not TempOK Then
      Begin  H := 0;  M := 0;  S := 0;  End;
    ParseTime := TempOK;
  End;



{----------------------------------------------------------------------}
{ Function AbsDate                                                     }
{                                                                      }
{ Given individual year, month, and day values, returns the number of  }
{ days since January 1st, 1979.  ie. 01-02-79 = 1.                     }
{ This function calculates the absolute date for all later functions.  }
{----------------------------------------------------------------------}

Function AbsDate( Y, M, D : Integer ) : LongInt;
  VAR  MLoop : Integer;  Total : LongInt;
  BEGIN
    Y := Y - 1980;
  { start with leap year days }
    Total := Y div 4;
  { compensate for fact that the base year, 1980, is a leap year }
    Inc( Total );
  { compensate if current year is a leap year, but leap day is not past }
    If ( Y mod 4 = 0 ) and
       ( ( M < 2 ) or ( ( M = 2 ) and ( D < 29 ) ) ) Then Dec( Total );
  { add days for past years }
    Total := Total + ( ( Y - 1 ) * 365 );
    MLoop := 1;
  { add days for elapsed months in year }
    While MLoop < M do
      BEGIN
        Total := Total + DaysPerMonth[ MLoop ];
        Inc( MLoop );
      END;
  { add elapsed days in month to finish }
    AbsDate := Total + D - 1;
  END;



{----------------------------------------------------------------------}
{ Function DateInDays                                                  }
{                                                                      }
{ returns system clock date in absolute format                         }
{----------------------------------------------------------------------}

Function DateInDays : LongInt;
  VAR  Y, M, D, Day : Word;
  BEGIN
    GetDate( Y, M, D, Day );
    DateInDays := AbsDate( Y, M, D );
  END;



{----------------------------------------------------------------------}
{ Function OrdDate                                                     }
{                                                                      }
{ converts a date string into absolute format                          }
{----------------------------------------------------------------------}

Function OrdDate( InDate : String8 ) : LongInt;

  var
    Y, M, D : Integer;

  Begin
    OrdDate := 0;
    If ParseDate( InDate, Y, M, D ) Then
      OrdDate := AbsDate( Y, M, D );
  End;



{----------------------------------------------------------------------}
{ Function StrDate                                                     }
{                                                                      }
{ converts an absolute date into string format                         }
{----------------------------------------------------------------------}

Function StrDate( InDate : LongInt ) : String8;

  var
    Y, M, D : Integer;

  Begin
    Y := 1980;  M := 1;
    While InDate > DaysPerMonth[ M ] do
      Begin
        InDate := InDate - DaysPerMonth[ M ];
        Inc( M );
        If M > 12 Then
          Begin  M := 1;  Inc( Y );  End;
      End;
    StrDate := MakeDate( Y, M, D );
  End;


{----------------------------------------------------------------------}
{ Function EqualDate                                                   }
{                                                                      }
{ returns true if two dates are equal                                  }
{----------------------------------------------------------------------}

Function EqualDate( InDate1, InDate2 : String8 ) : Boolean;
  BEGIN  EqualDate := OrdDate( InDate1 ) = OrdDate( InDate2 );  END;



{----------------------------------------------------------------------}
{ Function GreaterDate                                                 }
{                                                                      }
{ returns true if first date is later than second                      }
{----------------------------------------------------------------------}

Function GreaterDate( InDate1, InDate2 : String8 ) : Boolean;
  BEGIN  GreaterDate := OrdDate( InDate1 ) > OrdDate( InDate2 );  END;



{----------------------------------------------------------------------}
{ Function AbsTime                                                     }
{                                                                      }
{ Given individual hours, minutes, and seconds values, returns the     }
{ number of seconds since midnight.  ie.  00:00:01 = 1.                }
{ This function calculates the absolute time for all later functions.  }
{----------------------------------------------------------------------}

Function AbsTime( H, M, S : Integer ) : LongInt;  { seconds since 00:00:00 }
  BEGIN  AbsTime := LongInt( H ) * 3600 + M * 60 + S;  END;



{----------------------------------------------------------------------}
{ Function TimeInSeconds                                               }
{                                                                      }
{ returns system clock time in absolute format                         }
{----------------------------------------------------------------------}

Function TimeInSeconds : LongInt;
  VAR  H, M, S, S100 : Word;
  BEGIN
    GetTime( H, M, S, S100 );
    TimeInSeconds := AbsTime( H, M, S );
  END;



{----------------------------------------------------------------------}
{ Function OrdTime                                                     }
{                                                                      }
{ converts a time string into absolute format                          }
{----------------------------------------------------------------------}

Function OrdTime( InTime : String8 ) : LongInt;
  VAR  H, M, S : Integer;
  BEGIN
    OrdTime := 0;
    If not ParseTime( InTime, H, M, S ) THEN Exit;
    OrdTime := AbsTime( H, M, S );
  END;



{----------------------------------------------------------------------}
{ Function StrTime                                                     }
{                                                                      }
{ converts an absolute time into string format                         }
{----------------------------------------------------------------------}

Function StrTime( InTime : LongInt ) : String8;
  VAR  H, M, S : LongInt;
  BEGIN
    H := InTime div ( 60 { minutes } * 60 { seconds } );
    InTime := InTime - ( H * 60 { minutes } * 60 { seconds } );
    M := InTime div 60 { seconds };
    InTime := InTime - ( M * 60 { seconds } );
    S := InTime;
    StrTime := Str2( H ) + ':' + Str2( M ) + ':' + Str2( S );
  END;



{----------------------------------------------------------------------}
{ Function AbsDateTime                                                 }
{                                                                      }
{ Given individual year, month, date, hours, minutes, and seconds      }
{ values, returns number of seconds since 00:00:00 on 01-01-1980.      }
{----------------------------------------------------------------------}

Function AbsDateTime( Y, Month, D, H, Minute, S : Integer ) : Real;
  BEGIN
    AbsDateTime := OrdDate( MakeDate( Y, Month, D ) ) * 24 * 60 * 60.0
                   + OrdTime( MakeDate( H, Minute, S ) );
  END;



{----------------------------------------------------------------------}
{ Function DateTimeInSeconds                                           }
{                                                                      }
{ Returns the system clock date and time in absolute format            }
{ as seconds since 00:00:00 on 01-01-1980.                             }
{----------------------------------------------------------------------}

Function DateTimeInSeconds : Real;
  BEGIN
    DateTimeInSeconds := DateInDays * 24 * 60 * 60.0 + TimeInSeconds;
  END;



{----------------------------------------------------------------------}
{ Function ValidDate                                                   }
{                                                                      }
{ return True if string represents a valid date                        }
{----------------------------------------------------------------------}

Function ValidDate( InDate : String8 ) : Boolean;

  var
    Y, M, D : Integer;

  Begin
    ValidDate := False;
    If ParseDate( InDate, Y, M, D ) Then
      If ( M < 255 ) and ( D < 255 ) and
         ( Y <= Sys.MaxYear ) and ( Y >= Sys.MinYear ) Then
        ValidDate := ( M in [ 1 .. 12 ] ) and
                     ( ( D in [ 1 .. DaysPerMonth[ M ] ] ) or
                       ( ( D = 29 ) and ( M = 2 ) and ( Y mod 4 = 0 ) ) );
  End;



{----------------------------------------------------------------------}
{ Function ValidTime                                                   }
{                                                                      }
{ return True if string represents a valid time                        }
{----------------------------------------------------------------------}

Function ValidTime( InTime : String8 ) : Boolean;
  VAR  Error, H, M, S : Integer;
  BEGIN
    ValidTime := ParseTime( InTime, H, M, S ) and
                 ( H >= 0 ) and ( H <= 23 ) and
                 ( M >= 0 ) and ( M <= 59 ) and
                 ( S >= 0 ) and ( S <= 59 );
  END;




{----------------------------------------------------------------------}
{ Function NextDate                                                    }
{                                                                      }
{ return the next calendar date given a date string                    }
{----------------------------------------------------------------------}

Function NextDate( InDate : String8 ) : String8;
  CONST   { re-declared here for self-modification }
    DaysInMonth : Array [ 1..12 ] of Integer =
           ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
           {  J   F   M   A   M   J   J   A   S   O   N   D }
  VAR  Y, M, D : integer;
  Begin
    InDate := Date6( InDate );  NextDate := InDate;
    If not ParseDate( InDate, Y, M, D ) THEN Exit;
  { Calculate if this is a leap year, change feb's entry if so}
    If Y mod 4 = 0  { if leap year}
      THEN DaysInMonth[2] := 29   { leap year   }
      ELSE DaysInMonth[2] := 28;  { normal Year }
    Inc( D ); { bump day }
    If D > DaysInMonth[ M ]    { bump month }
      THEN BEGIN
        D := 1;  Inc( M );
        If M > 12 { bump year }
          THEN BEGIN  Inc( Y );  M := 1;  END;
      END;
    NextDate := MakeDate( Y, M, D );
  End;



{----------------------------------------------------------------------}
{ Function PrevDate                                                    }
{                                                                      }
{ return the previous calendar date given a date string                }
{----------------------------------------------------------------------}

Function PrevDate( InDate : String8 ) : String8;
  CONST   { re-declared here for self-modification }
    DaysInMonth : Array [ 1..12 ] of Integer =
           ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
           {  J   F   M   A   M   J   J   A   S   O   N   D }
  VAR  Y, M, D : integer;
  Begin
    InDate := Date6( InDate );  PrevDate := InDate;
    If not ParseDate( InDate, Y, M, D ) THEN Exit;
  { Calculate if this is a leap year, change feb's entry if so}
    If Y mod 4 = 0  { if leap year}
      THEN DaysInMonth[2] := 29   { leap year   }
      ELSE DaysInMonth[2] := 28;  { normal Year }
    Dec( D ); { dec day }
    If D < 1
      THEN BEGIN
        Dec( M );    { dec month }
        If M < 1  { dec year }
          THEN BEGIN  Dec( Y );  M := 12;  END;
        D := DaysInMonth[ M ];
      END;
    PrevDate := MakeDate( Y, M, D );
  END;



{----------------------------------------------------------------------}
{ Function YYMMDD                                                      }
{                                                                      }
{ swap an MM-DD-YY date into YY-MM-DD format                           }
{----------------------------------------------------------------------}

Function YYMMDD( InDate : String8 ) : String8;
  Begin
    InDate := Date6( InDate );
    InDate := Copy( InDate, 5, 2 ) + Copy( InDate, 1, 4 );
    YYMMDD := Date8( InDate );
  End;



{----------------------------------------------------------------------}
{ Function DateDay                                                     }
{                                                                      }
{ return a day number given a date string                              }
{----------------------------------------------------------------------}

Function DateDay( InDate : String8 ) : Byte;
  BEGIN
    DateDay := (OrdDate( InDate ) - 4)  { 01-01-80 (Tuesday) = 0 }
               mod 7    { modulo into weeks }
               + 1;     { offset for Tuesday }
  END;



{----------------------------------------------------------------------}
{ Function Day                                                         }
{                                                                      }
{ return a day name given a date string                                }
{----------------------------------------------------------------------}

Function Day( InDate : String8 ) : SmallString;
  BEGIN  Day := DaysOfWeek[ DateDay( InDate ) ];  END;



{----------------------------------------------------------------------}
{ Function ShowDate                                                    }
{                                                                      }
{ return the full date in a selection of high level formats            }
{----------------------------------------------------------------------}

Function ShowDate( InDate : String8;  Style : Byte ) : String40;

  { outputs date formatted in a choice of styles:  }
  {   0 = '01-01-80'  or European : 'DD-MM-YY'     }
  {   1 = 'Jan. 1, 1980'                           }
  {   2 = 'Tue., Jan. 1, 1980'                     }
  {   3 = 'Tuesday, January 1st, 1980'             }
  { Months of four letters are not abreviated.     }

  VAR  Y, M, D : Integer;
       ShowDay, ShowMonth, ShowMDate, ShowExt, ShowYear : SmallString;

  BEGIN
    If ( Style > 3 ) or ( Style < 0) THEN
      Style := 0;
    If not ParseDate( InDate, Y, M, D ) THEN
      Begin
        ShowDate := 'Bad date';
        Exit;
      End;
    Case D of
      1 : ShowExt := 'st';
      2 : ShowExt := 'nd';
      3 : ShowExt := 'rd';
      ELSE ShowExt := 'th';
    End;
    ShowYear := I_S( Y );      { show year }
    ShowMonth := Months[ M ];  { show month }
    ShowMDate := I_S( D );     { show date }
    ShowDay := Day( InDate );  { show day }
    Case Style of
      1 : BEGIN
            If Length( ShowMonth ) > 4
              THEN ShowMonth := Abrv( ShowMonth, 4 );
            ShowDay := '';
            ShowDate := ShowDay + ShowMonth + ' ' +
                        ShowMDate + ', ' + ShowYear;
          END;
      2 : BEGIN
            If Length( ShowMonth ) > 4
              THEN ShowMonth := Abrv( ShowMonth, 4 );
            ShowDay := Abrv( ShowDay, 4 ) + ', ';
            ShowDate := ShowDay + ShowMonth + ' ' +
                        ShowMDate + ', ' + ShowYear;
          END;
      3:  BEGIN
            ShowDay := ShowDay + ', ';
            ShowMDate := ShowMDate + ShowExt + '.';
            ShowDate := ShowDay + ShowMonth + ' ' +
                        ShowMDate + ', ' + ShowYear;
          END;
      0:  BEGIN
            If Sys.EuropeanDate Then  { use DDMMYY rather than MMDDYY }
              InDate := Copy(InDate,4,2) + Copy(InDate,1,2) +
                        Copy(InDate,7,2);
            ShowDate := Date8(InDate);
          END;
    End;
  END;



{----------------------------------------------------------------------}
{ Procedure SetTimer                                                   }
{                                                                      }
{ This is the interface to the interrupt serviced timer maintenance    }
{ procedure.  Specify a timer number [1..5] and a duration in          }
{ milliseconds.  The TimerElapsed procedure will then return true      }
{ when the specified time has elapsed.  Accuracy is 1/18.2 seconds.    }
{ The showtime procedure uses timer zero.                              }
{----------------------------------------------------------------------}

Procedure SetTimer( TimerNum : Integer;  Duration : LongInt );
  VAR  TimeCalc : LongInt;
  BEGIN
    If ( TimerNum < MinTimer ) or ( TimerNum > MaxTimer ) then Exit;
    TimeCalc :=
      Trunc( ( ( Duration + 0.0 ) / 1000.0 ) * 18.2 );
    If TimeCalc > 65535 then TimeCalc := 65535;
    Timer[ TimerNum ] := TimeCalc;
  END;



{----------------------------------------------------------------------}
{ Procedure TimerElapsed                                               }
{                                                                      }
{ Returns true when requested background timer has elapsed.            }
{ See SetTimer procedure above.                                        }
{----------------------------------------------------------------------}

Function TimerElapsed( TimerNum : Integer ) : Boolean;
  BEGIN
    TimerElapsed :=
      ( TimerNum < MinTimer ) or ( TimerNum > MaxTimer ) or
      ( Timer[ TimerNum ] = 0 );
  END;



{----------------------------------------------------------------------}
{ Procedure ServiceTimers                                              }
{                                                                      }
{ Resident timer service routine.  See SetTimer procedure above.       }
{----------------------------------------------------------------------}

{$R-,S-}
Procedure ServiceTimers( flags,cs,ip,ax,bx,cx,dx,si,di,ds,es,bp : Word );
  Interrupt;
  BEGIN
    InLine( $FA );
    For ISRLoop := MinTimer to MaxTimer do
      If Timer[ ISRLoop ] <> 0
        THEN Dec( Timer[ ISRLoop ] );
    { Serves one second background operations, used for clock display. }
    If Timer[ ClockTimer ] = 0
      THEN BEGIN  { second elapsed }
        Timer[ ClockTimer ] := SecondServiceFreq;  { reset one second countdown }
        If Sys.ShowTimeEnabled and ( not Sys.RemoteOnly )
          THEN BEGIN
          { display the time without the help of the video module }
            ISRHours := Mem[ $40 : $6E ];   { hours }
            ISRMinutes := MemW[ $40 : $6C ] div 1092;  { minutes }
            ISRSeconds := MemW[ $40 : $6C ] mod 1092 div 18;  { sec }
            Screen^[ ScreenHeight, 72 ].Ch :=
              Chr( ISRHours div 10 + Ord( '0' ) );
            Screen^[ ScreenHeight, 73 ].Ch :=
              Chr( ISRHours mod 10 + Ord( '0' ) );
            Screen^[ ScreenHeight, 74 ].Ch := ':';
            Screen^[ ScreenHeight, 75 ].Ch :=
              Chr( ISRMinutes div 10 + Ord( '0' ) );
            Screen^[ ScreenHeight, 76 ].Ch :=
              Chr( ISRMinutes mod 10 + Ord( '0' ) );
            Screen^[ ScreenHeight, 77 ].Ch := ':';
            Screen^[ ScreenHeight, 78 ].Ch :=
              Chr( ISRSeconds div 10 + Ord( '0' ) );
            Screen^[ ScreenHeight, 79 ].Ch :=
              Chr( ISRSeconds mod 10 + Ord( '0' ) );
          END;  { ShowTimeEnabled }
      END;  { second elapsed }
    InLine( $CD/FixedTimerVector/$FB );  { re-enabled interrupts }
  END;
{$R+,S+}

{----------------------------------------------------------------------}
{ Procedure ReleaseTimers                                              }
{                                                                      }
{ Restores timer tick interrupt vector to operating system.            }
{----------------------------------------------------------------------}

Procedure ReleaseTimers;
  BEGIN
    SetIntVec( $1C, TimerOriginalAddress );
    SetIntVec( TimerNewVector, NIL );
  { this clears any leftover time display }
    For ISRLoop := 71 to 79 do
      Screen^[ ScreenHeight, ISRLoop ].Ch := ' ';
    ExitProc := TimeExitSave;
    Halt;
  END;



{-----------------------------------------------------------------------}

BEGIN
  Sys.MinYear := 1980;  Sys.MaxYear := 2079;
  FillChar( Timer, SizeOf( Timer ), 0 );
  TimeExitSave := ExitProc;
  ExitProc := @ReleaseTimers;           { prepare to restore tick interrupt }
  TimerNewVector := FixedTimerVector { FreeVector };    { displace timer }
  GetIntVec( $1C, TimerOriginalAddress );  { preserve timer tick interrupt }
  SetIntVec( TimerNewVector, TimerOriginalAddress );
  SetIntVec( $1C, @ServiceTimers );     { displace timer tick interrupt }
END.



